<?php

/**
 * ownCloud
 *
 * @author Artur Neumann <artur@jankaritech.com>
 * @copyright Copyright (c) 2017 Artur Neumann artur@jankaritech.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU AFFERO GENERAL PUBLIC LICENSE for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

namespace Page;

use Behat\Mink\Element\NodeElement;
use Behat\Mink\Session;
use SensioLabs\Behat\PageObjectExtension\PageObject\Exception\ElementNotFoundException;
use WebDriver\Exception\StaleElementReference;

/**
 * PageObject for the FileFirewall Page
 *
 */
class FileFirewallPage extends OwncloudPage {

	/**
	 *
	 * @var string $path
	 */
	protected $path = '/index.php/settings/admin?sectionid=security';
	protected $groupContainerXpath = './/*[@class="rules-group-container"]';
	protected $groupNameInputXpath = './/input[@name="groupName"]';
	protected $groupContainerFromNameInputXpath = "/../../..";
	protected $saveRulesBtnXpath = ".//button[@class='saveRules']";
	protected $addGroupBtnXpath = ".//button[@data-add='group']";
	protected $logLevelSelectId = "firewallDebug";

	/**
	 * finds a firewall group by its name and returns a FileFirewallGroup Page Object
	 *
	 * @param string $name
	 *
	 * @throws ElementNotFoundException
	 * @return FileFirewallGroup
	 */
	public function findGroupByName($name) {
		$currentTime = \microtime(true);
		$end = $currentTime + (STANDARD_UI_WAIT_TIMEOUT_MILLISEC / 1000);
		while ($currentTime <= $end) {
			try {
				$groupNameInputs = $this->findAll(
					"xpath", $this->groupNameInputXpath
				);

				if ($groupNameInputs === null) {
					throw new ElementNotFoundException(
						"could not find any firewall group name input fields"
					);
				}

				foreach ($groupNameInputs as $groupNameInput) {
					if ($groupNameInput->getValue() === $name) {
						$groupContainer = $groupNameInput->find(
							"xpath",
							$this->groupContainerFromNameInputXpath
						);

						if ($groupContainer === null) {
							throw new ElementNotFoundException(
								"could not find group container"
							);
						}

						$fileFirewallGroup = $this->getPage("FileFirewallGroup");
						$fileFirewallGroup->setElement($groupContainer);
						return $fileFirewallGroup;
					}
				}
			} catch (StaleElementReference $e) {
				\usleep(STANDARD_SLEEP_TIME_MICROSEC);
			}
			$currentTime = \microtime(true);
		}
		throw new ElementNotFoundException("could not find group named $name");
	}

	/**
	 * finds a firewall group by its number and returns a
	 * FileFirewallGroup Page Object
	 *
	 * @param int $number counting of elements starts from 1.
	 *                    negative numbers return elements counted from the end.
	 *                    -1 last element -2 second to last etc
	 *
	 * @throws ElementNotFoundException
	 * @return FileFirewallGroup
	 */
	public function findGroupByNumber($number) {
		/**
		 * @var FileFirewallGroup $fileFirewallGroup
		 */
		$fileFirewallGroup = $this->getPage("FileFirewallGroup");

		$number = (int) $number;
		if ($number < 0) {
			$number++;
			if ($number === 0) {
				$number = "last()";
			} else {
				$number = "last()$number";
			}
		}
		$groupContainer = $this->find(
			"xpath",
			$this->groupContainerXpath . "[$number]"
		);

		if ($groupContainer === null) {
			throw new ElementNotFoundException(
				"could not find firewall group with the number $number"
			);
		}
		$fileFirewallGroup->setElement($groupContainer);
		return $fileFirewallGroup;
	}

	/**
	 * finds and returns all group containers
	 *
	 * @throws ElementNotFoundException
	 * @return NodeElement[]
	 */
	public function getAllGroupContainers() {
		$containers = $this->findAll("xpath", $this->groupContainerXpath);
		if ($containers === null) {
			throw new ElementNotFoundException("could not find any groups on page");
		}
		return $containers;
	}

	/**
	 * adds a new firewall group and returns a FileFirewallGroup Page Object
	 *
	 * @param int $timeout_msec
	 *
	 * @return FileFirewallGroup
	 * @throws ElementNotFoundException
	 * @throws \Exception
	 */
	public function addGroup($timeout_msec = STANDARD_UI_WAIT_TIMEOUT_MILLISEC) {
		$addGroupBtn = $this->find("xpath", $this->addGroupBtnXpath);
		if ($addGroupBtn === null) {
			throw new ElementNotFoundException(
				"could not find add-group-button on page"
			);
		}
		$oldGroupCount = \count($this->getAllGroupContainers());
		
		$addGroupBtn->click();

		$currentTime = \microtime(true);
		$end = $currentTime + ($timeout_msec / 1000);
		while ($currentTime <= $end) {
			$allGroups = $this->getAllGroupContainers();
			if (\count($allGroups) > $oldGroupCount) {
				break;
			}
			\usleep(STANDARD_SLEEP_TIME_MICROSEC);
			$currentTime = \microtime(true);
		}
		if (\count($allGroups) <= $oldGroupCount) {
			throw new \Exception("could not add a new group");
		}
		
		$fileFirewallGroup = $this->findGroupByNumber(-1);
		return $fileFirewallGroup;
	}

	/**
	 * saves all rules on the page
	 *
	 * @param Session $session
	 *
	 * @throws ElementNotFoundException
	 * @return void
	 */
	public function saveRules(Session $session) {
		$saveRulesBtn = $this->find("xpath", $this->saveRulesBtnXpath);
		if ($saveRulesBtn === null) {
			throw new ElementNotFoundException(
				"could not find save-button on page"
			);
		}
		$saveRulesBtn->click();
		$this->waitForOutstandingAjaxCalls($session);
	}

	/**
	 *
	 * @param string $logLevel
	 * @param Session $session
	 *
	 * @return void
	 */
	public function setLogLevel($logLevel, Session $session) {
		$this->selectFieldOption($this->logLevelSelectId, $logLevel);
		$this->waitForOutstandingAjaxCalls($session);
	}
}
