<?php
/**
 * ownCloud Firewall
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright Copyright (c) 2018, ownCloud GmbH
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Firewall\Tests\Lib;

use OCA\Firewall\Config;
use OCA\Firewall\Context;
use OCP\Files\IMimeTypeDetector;
use OCP\Files\IRootFolder;
use OCP\IGroupManager;
use OCP\IRequest;
use OCP\IUserSession;
use PHPUnit\Framework\MockObject\MockObject;
use Sabre\DAV\Server;
use Sabre\DAV\Tree;
use Test\TestCase;

class ContextTest extends TestCase {

	/** @var IGroupManager|MockObject */
	protected $groupManager;

	/** @var IRequest|MockObject */
	protected $request;

	/** @var IUserSession|MockObject */
	protected $userSession;

	/** @var IMimeTypeDetector|MockObject */
	protected $mimeTypeDetector;

	/** @var Config|MockObject */
	protected $config;

	/**
	 *
	 * @var IRootFolder|MockObject
	 */
	protected $rootFolder;

	protected $sabreServer;

	protected $sabreTree;

	protected function setUp(): void {
		parent::setUp();

		$this->groupManager = $this->getMockBuilder(IGroupManager::class)
			->disableOriginalConstructor()
			->getMock();

		$this->request = $this->getMockBuilder(IRequest::class)
			->disableOriginalConstructor()
			->getMock();

		$this->userSession = $this->getMockBuilder(IUserSession::class)
			->disableOriginalConstructor()
			->getMock();

		$this->mimeTypeDetector = $this->getMockBuilder(IMimeTypeDetector::class)
			->disableOriginalConstructor()
			->getMock();

		$this->config = $this->getMockBuilder(Config::class)
			->disableOriginalConstructor()
			->getMock();

		$this->rootFolder = $this->getMockBuilder(IRootFolder::class)
			->disableOriginalConstructor()
			->getMock();

		$this->sabreServer = $this->getMockBuilder(Server::class)
			->disableOriginalConstructor()
			->getMock();

		$this->sabreTree = $this->getMockBuilder(Tree::class)
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return Context|MockObject
	 */
	protected function getInstance($methods = []) {
		if (empty($methods)) {
			return new Context(
				$this->request,
				$this->groupManager,
				$this->userSession,
				$this->mimeTypeDetector,
				$this->config,
				$this->rootFolder
			);
		} else {
			return $this->getMockBuilder(Context::class)
				->setConstructorArgs(
					[
						$this->request,
						$this->groupManager,
						$this->userSession,
						$this->mimeTypeDetector,
						$this->config,
						$this->rootFolder
					]
				)
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataGetRequestFullUrl() {
		return [
			['http', 'localhost', '/', 'http://localhost/'],
			['https', 'owncloud.org', '/index.php', 'https://owncloud.org/index.php'],
			['https', 'localhost', '/owncloud/index.php', 'https://localhost/owncloud/index.php'],
		];
	}

	/**
	 * @dataProvider dataGetRequestFullUrl
	 *
	 * @param string $protocol
	 * @param string $host
	 * @param string $requestUri
	 * @param string $expected
	 */
	public function testGetRequestFullUrl($protocol, $host, $requestUri, $expected) {
		$context = $this->getInstance();

		$this->request->expects($this->once())
			->method('getServerProtocol')
			->willReturn($protocol);

		$this->request->expects($this->once())
			->method('getServerHost')
			->willReturn($host);

		$this->request->expects($this->once())
			->method('getRequestUri')
			->willReturn($requestUri);

		$this->assertSame($expected, $context->getRequestFullUrl());
	}

	public function dataGetRequestUri() {
		return [
			['/'],
			['/index.php'],
			['/owncloud/index.php'],
		];
	}

	/**
	 * @dataProvider dataGetRequestUri
	 *
	 * @param string $requestUri
	 */
	public function testGetRequestUri($requestUri) {
		$context = $this->getInstance();

		$this->request->expects($this->once())
			->method('getRequestUri')
			->willReturn($requestUri);

		$this->assertSame($requestUri, $context->getRequestUri());
	}

	public function dataGetRequestType() {
		return [
			['/index.php', '/apps/files', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/index.php', '/apps/files', true, 'user1', Context::REQUEST_TYPE_OTHER],
			['/public.php', '/webdav', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '/webdav', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '?service=files&token=tokenis13char', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '?service=files&token=tokenis13char', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/index.php', '/s/tokenis13char', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/index.php', '/s/tokenis13char', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/remote.php', '/webdav', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/some/path', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/some/path', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/user1/path', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/user1/path', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/uploads/', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/uploads/', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/uploads/user1/path', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/uploads/user1/path', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/addressbooks', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks', true, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks/', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks/', true, 'user1', Context::REQUEST_TYPE_OTHER],

			['/index.php', '/apps/files_drop/tokenis13char', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/index.php', '/apps/files_drop/tokenis13char', true, 'user1', Context::REQUEST_TYPE_FILES_DROP],
			['/index.php', '/apps/files_sharing/ajax/publicpreview.php', false, null, Context::REQUEST_TYPE_PUBLIC],
		];
	}

	/**
	 * @dataProvider dataGetRequestType
	 *
	 * @param string $scriptName
	 * @param string $pathInfo
	 * @param bool $verbose
	 * @param string|null $sessionUser
	 * @param string $expected
	 */
	public function testGetRequestType($scriptName, $pathInfo, $verbose, $sessionUser, $expected) {
		$context = $this->getInstance();

		if ($sessionUser === null) {
			$this->userSession->expects($this->any())
				->method('getUser')
				->willReturn(null);
		} else {
			$this->userSession->expects($this->any())
				->method('getUser')
				->willReturn($this->createMock('OCP\IUser'));
		}

		$this->request->expects($this->atLeastOnce())
			->method('getScriptName')
			->willReturn($scriptName);

		$this->request->expects($this->any())
			->method('getPathInfo')
			->willReturn($pathInfo);

		$this->assertSame($expected, $context->getRequestType($verbose));
	}

	public function testGetRequestTime() {
		$context = $this->getInstance();

		$before = \time();
		$actual = $context->getRequestTime();
		$after = \time();

		$this->assertGreaterThanOrEqual($before, $actual);
		$this->assertLessThanOrEqual($after, $actual);
	}

	public function dataGetUploadSize() {
		return [
			[2, null, null, null],
			[2, 0, null, null],
			[2, 1, null, null],
			[2, 0, 1, 1],
			[2, 1, 0, 0],
			[2, null, 0, 0],
			[2, null, 1, 1],
		];
	}

	/**
	 * @dataProvider dataGetUploadSize
	 *
	 * @param int $headerCalls
	 * @param int $ocTotalLength
	 * @param int $contentLength
	 * @param int $expected
	 */
	public function testGetUploadSizeSizeHandling($headerCalls, $ocTotalLength, $contentLength, $expected) {
		$context = $this->getInstance([
			'getRequestType'
		]);

		$context->expects($this->once())
			->method('getRequestType')
			->willReturn(Context::REQUEST_TYPE_WEBDAV);

		$this->request->expects($this->once())
			->method('getMethod')
			->willReturn('PUT');

		$this->request->expects($this->exactly($headerCalls))
			->method('getHeader')
			->willReturnMap([
				['OC_TOTAL_LENGTH', $ocTotalLength],
				['CONTENT_LENGTH', $contentLength],
			]);

		$this->assertSame($expected, $context->getUploadSize());
	}

	/**
	 * data provider for testGetUploadSizeCheck
	 *
	 * @return array see testGetUploadSizeCheck() for explanation
	 */
	public function dataGetUploadSizeCheck() {
		return [
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', null, 10, 10, 0, null, 10],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['not_empty'], 100, null, null, null, 100],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['not_empty'], 100, 10, null, null, 100],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/webdav/file1-chunking-3-2'], 100, null, [60, 40], "v1", 200],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/webdav/file1-chunking-3-2'], 100, 10, [60, 40], "v1", 200],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/dav/uploads/admin/111/2'], 100, null, [60, 40], "v2", 200],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/dav/uploads/admin/111/2'], 100, 10, [60, 40], "v2", 200],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/dav/uploads/admin/111/1'], 100, null, [60, 40, 30], "v2", 190],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', ['/dav/uploads/admin/111/1'], 100, 10, [60, 40, 30], "v2", 190],
			[Context::REQUEST_TYPE_WEBDAV, 'MKCOL', ['/dav/uploads/admin/111'], 10, 100, [0], "v2", 100],
			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', ['/dav/uploads/admin/111/.file'], null, null, [60, 40, 30], "v2", 130],
			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', ['/dav/uploads/admin/111/.file'], 0, null, [60, 40, 30], "v2", 130],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', null, 10, 10, 0, null, null],

			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', ['/webdav/file1'], 10, 10, 0, null, 10],
			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', ['/webdav/file1'], null, null, 110, null, 110],
			[Context::REQUEST_TYPE_WEBDAV, 'COPY', ['/webdav/file1'], null, null, 120, null, 120],
			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', ['/dav/files/admin/1.JPG'], null, null, 110, null, 110],
			[Context::REQUEST_TYPE_WEBDAV, 'COPY', ['/dav/files/admin/1.JPG'], null, null, 120, null, 120],

			[Context::REQUEST_TYPE_PUBLIC, 'PUT', ['/webdav/file1'], 100, null, null, null, 100],

			[Context::REQUEST_TYPE_OTHER, 'GET', null, 10, 10, 0, null, null],
			[Context::REQUEST_TYPE_OTHER, 'POST', [], 10, 10, 0, null, null],
			[Context::REQUEST_TYPE_OTHER, 'POST', ['not_empty'], 10, 10, 0, null, 10],
			[Context::REQUEST_TYPE_OTHER, 'PUT', [], 10, 10, 0, null, null],
			[Context::REQUEST_TYPE_OTHER, 'PUT', ['not_empty'], 10, 10, 0, null, 10],

			[Context::REQUEST_TYPE_FILES_DROP, 'PUT', null, 10, 10, 0, null, null],
			[Context::REQUEST_TYPE_FILES_DROP, 'POST', [], 10, 10, 0, null, null],
			[Context::REQUEST_TYPE_FILES_DROP, 'POST', ['not_empty'], 10, 10, 0, null, 10],
		];
	}

	/**
	 * @dataProvider dataGetUploadSizeCheck
	 *
	 * @param string $requestType
	 * @param string $requestMethod
	 * @param array|null $uploadedFiles
	 * @param int $contentLength Content-Length HTTP Header
	 * @param int $ocTotalLength OC-Total-Length HTTP Header
	 * @param array $alreadyUploadedSize size of already uploaded data.
	 * Can be size of sourcefile in MOVE/COPY operation or
	 * sizes of chunks that were already uploaded
	 * @param string|null $chunkingVersion "v1|v2"
	 * @param bool $expected expected size
	 * @return void
	 */
	public function testGetUploadSizeCheck(
		$requestType, $requestMethod, $uploadedFiles, $contentLength,
		$ocTotalLength, $alreadyUploadedSize, $chunkingVersion, $expected
	) {
		$davPathVersion = null;
		if (isset($uploadedFiles[0])
			&& \substr($uploadedFiles[0], 0, \strlen('/dav/')) === '/dav/'
		) {
			$davPathVersion = 2;
		}
		$context = $this->getInstance(
			[
				'getRequestType',
				'isV1WebdavChunk',
				'createV1Chunkedfile',
				'decodeV1ChunkingName',
				'getFileCache',
				'getFileSizeFromFilesystem'
			]
		);

		$context->expects($this->atLeastOnce())
			->method('getRequestType')
			->willReturn($requestType);

		$this->request->expects($this->once())
			->method('getPathInfo')
			->willReturn(isset($uploadedFiles[0]) ? $uploadedFiles[0] : null);

		$this->request->expects($this->once())
			->method('getMethod')
			->willReturn($requestMethod);

		$this->request->expects($this->atLeastOnce())
			->method('getHeader')->will(
				$this->returnValueMap(
					[
						[
							'CONTENT_LENGTH',
							$contentLength
						],
						[
							'OC_TOTAL_LENGTH',
							$ocTotalLength
						]
					]
				)
			);

		if ($requestType === Context::REQUEST_TYPE_WEBDAV
			|| ($requestType === Context::REQUEST_TYPE_PUBLIC && $requestMethod === 'PUT')
		) {
			if ($chunkingVersion === "v1") {
				$context->expects($this->atLeastOnce())
					->method('isV1WebdavChunk')
					->willReturn(true);

				$indexParseArray = \explode("-", $uploadedFiles[0]);
				$currentChunkIndex = $indexParseArray[\sizeof($indexParseArray) - 1];
				$context->expects($this->atLeastOnce())
					->method('decodeV1ChunkingName')
					->willReturn(
						[
							'index' => $currentChunkIndex,
							'chunkcount' => \count($alreadyUploadedSize)
						]
					);

				$chunkedFile = $this->getMockBuilder('OC_FileChunking')
					->disableOriginalConstructor()
					->setMethods(['getPrefix'])
					->getMock();
				$chunkFileNamePrefix = \basename($uploadedFiles[0]) . "chunk-";
				$chunkedFile->expects($this->atLeastOnce())
					->method('getPrefix')
					->willReturn($chunkFileNamePrefix);

				$fileCache = $this->createMock('OC\Cache\File');
				$sizeValueMap = [];

				//sizes for every chunk
				for ($i = 0; $i < \sizeof($alreadyUploadedSize); $i++) {
					$sizeValueMap[] = [
						$chunkFileNamePrefix . $i, $alreadyUploadedSize[$i]
					];
				}
				$fileCache->expects($this->atLeastOnce())
					->method('size')
					->will($this->returnValueMap($sizeValueMap));

				$context->expects($this->atLeastOnce())
					->method('createV1Chunkedfile')
					->willReturn($chunkedFile);

				$context->expects($this->atLeastOnce())
					->method('getFileCache')
					->willReturn($fileCache);
			} elseif ($chunkingVersion === "v2" && $requestMethod !== 'MKCOL') {
				//mock all chunks
				for ($i = 0; $i < \sizeof($alreadyUploadedSize); $i++) {
					$child = $this->createMock('Sabre\DAV\IFile');
					$child->expects($this->any())
						->method('getSize')
						->willReturn($alreadyUploadedSize[$i]);
					$child->expects($this->atLeastOnce())
						->method('getName')
						->willReturn((string)$i);
					$children[] = $child;
				}

				//mock the future file
				$child = $this->createMock('OCA\DAV\Upload\FutureFile');
				$child->expects($this->never())
					->method('getSize');
				$this->sabreTree->expects($this->once())
					->method('getChildren')
					->willReturn($children);
				$this->sabreServer->tree = $this->sabreTree;
			} elseif ($requestMethod === 'MOVE' || $requestMethod === 'COPY') {
				if ($davPathVersion === 2) {
					$node = $this->createMock('Sabre\DAV\IFile');
					$node->expects($this->once())
						->method('getSize')
						->willReturn($alreadyUploadedSize);

					$this->sabreTree->expects($this->once())
						->method('getNodeForPath')
						->willReturn($node);
					$this->sabreServer->tree = $this->sabreTree;
				} else {
					$context->expects($this->once())
						->method('getFileSizeFromFilesystem')
						->willReturn($alreadyUploadedSize);
				}
			}
		} else {
			$this->request->expects(
				$uploadedFiles === null ? $this->never() : $this->once()
			)
				->method('getUploadedFile')
				->willReturn($uploadedFiles);
		}
		$this->assertSame(
			$expected,
			$context->getUploadSize(
				$davPathVersion === 2 ? $this->sabreServer : null
			)
		);
	}

	public function dataGetUploadType() {
		return [
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', 'test.txt', [], 'text/plain'],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', 'test.mp3', [], 'audio/mpeg'],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', '', [], null],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', 'test.mp3', [], null],
			[Context::REQUEST_TYPE_WEBDAV, 'MOVE', 'test.mp3', [], 'audio/mpeg'],

			[Context::REQUEST_TYPE_OTHER, 'GET', '', [], null],
			[Context::REQUEST_TYPE_OTHER, 'GET', '', [
				'type' => [
				],
			], null],
			[Context::REQUEST_TYPE_OTHER, 'GET', '', [
				'type' => [
					0 => 'text/plain'
				],
			], null],

			[Context::REQUEST_TYPE_OTHER, 'POST', '', [], null],
			[Context::REQUEST_TYPE_OTHER, 'POST', '', [
				'type' => [
				],
			], null],
			[Context::REQUEST_TYPE_OTHER, 'POST', '', [
				'name' => [
					0 => 'test.txt'
				],
			], 'text/plain'],
			[Context::REQUEST_TYPE_OTHER, 'PUT', '', [
				'name' => [
					0 => 'test.txt'
				],
			], 'text/plain'],

			[Context::REQUEST_TYPE_FILES_DROP, 'POST', '', [], null],
			[Context::REQUEST_TYPE_FILES_DROP, 'POST', '', [
				'type' => null,
			], null],
			[Context::REQUEST_TYPE_FILES_DROP, 'POST', '', [
				'name' => 'test.txt',
			], 'text/plain'],
			[Context::REQUEST_TYPE_FILES_DROP, 'PUT', '', [], null],
		];
	}

	/**
	 * @dataProvider dataGetUploadType
	 *
	 * @param string $requestType
	 * @param string $requestMethod
	 * @param string $webDavPath
	 * @param array $files
	 * @param mixed $expected
	 */
	public function testGetUploadType($requestType, $requestMethod, $webDavPath, $files, $expected) {
		$context = $this->getInstance([
			'getRequestType'
		]);
		$context->method('getRequestType')
			->willReturn($requestType);

		if ($requestType === Context::REQUEST_TYPE_WEBDAV) {
			$this->request->expects($this->any())
				->method('getMethod')
				->willReturn($requestMethod);

			if ($requestMethod === 'MOVE') {
				$this->request->expects($this->any())
					->method('getHeader')
					->with($this->equalTo('DESTINATION'))
					->willReturn($webDavPath);
				$this->request->expects($this->once())
					->method('getPathInfo')
					->willReturn('/dav/uploads/' . $webDavPath . '/.file');
			} else {
				$this->request->expects($requestMethod === 'PUT' ? $this->once() : $this->never())
					->method('getPathInfo')
					->willReturn($webDavPath);
			}

			$this->mimeTypeDetector->expects($this->any())
				->method('detectPath')
				->with($webDavPath)
				->willReturnMap([
					['test.txt', 'text/plain'],
					['test.mp3', 'audio/mpeg'],
				]);
		} else {
			$this->request->expects($this->once())
				->method('getMethod')
				->willReturn($requestMethod);

			if ($requestType === Context::REQUEST_TYPE_FILES_DROP) {
				$this->mimeTypeDetector->expects($this->any())
					->method('detectPath')
					->willReturnMap([
						['test.txt', 'text/plain'],
						['test.mp3', 'audio/mpeg'],
					]);

				$this->request->expects($requestMethod === 'POST' ? $this->once() : $this->never())
					->method('getUploadedFile')
					->with('file')
					->willReturn($files);
			} else {
				$this->mimeTypeDetector->expects($this->any())
					->method('detectPath')
					->willReturnMap([
						['test.txt', 'text/plain'],
						['test.mp3', 'audio/mpeg'],
					]);

				$this->request->expects(\in_array($requestMethod, ['POST', 'PUT']) ? $this->once() : $this->never())
					->method('getUploadedFile')
					->with('files')
					->willReturn($files);
			}
		}

		$this->assertSame($expected, $context->getUploadType());
	}

	public function dataGetUserGroups() {
		$user = $this->getMockBuilder('OCP\IUser')
			->disableOriginalConstructor()
			->getMock();

		return [
			[null, []],
			[$user, []],
			[$user, ['admin']],
			[$user, ['admin', 'test1']],
		];
	}

	/**
	 * @dataProvider dataGetUserGroups
	 *
	 * @param mixed $user
	 * @param array $expected
	 */
	public function testGetUserGroups($user, array $expected) {
		$context = $this->getInstance();

		$this->userSession->expects($this->once())
			->method('getUser')
			->willReturn($user);

		if ($user !== null) {
			$this->groupManager->expects($this->once())
				->method('getUserGroupIds')
				->with($user)
				->willReturn($expected);
		} else {
			$this->groupManager->expects($this->never())
				->method('getUserGroupIds');
		}

		$this->assertSame($expected, $context->getUserGroups());
	}

	public function dataGetRemoteAddress() {
		return [
			['127.0.0.1'],
			['192.168.2.100'],
		];
	}

	/**
	 * @dataProvider dataGetRemoteAddress
	 *
	 * @param string $remoteAddress
	 */
	public function testGetRemoteAddress($remoteAddress) {
		$context = $this->getInstance();

		$this->request->expects($this->once())
			->method('getRemoteAddress')
			->willReturn($remoteAddress);

		$this->assertSame($remoteAddress, $context->getRemoteAddress());
	}

	public function dataGetRemoteServerAddress() {
		return [
			['127.0.0.1', '127.0.0.1'],
			['', ''],
			[null, ''],
		];
	}

	/**
	 * @dataProvider dataGetRemoteServerAddress
	 *
	 * @param string $remoteServerAddress
	 * @param string $expected
	 */
	public function testGetRemoteServerAddress($remoteServerAddress, $expected) {
		$context = $this->getInstance();

		$this->request->expects($this->once())
			->method('getHeader')
			->with('SERVER_ADDR')
			->willReturn($remoteServerAddress);

		$this->assertSame($expected, $context->getRemoteServerAddress());
	}

	public function dataGetUserAgent() {
		return [
			['Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:40.0) Gecko/20100101 Firefox/40.0'],
			['Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/45.0.2454.85 Chrome/45.0.2454.85 Safari/537.36'],
		];
	}

	/**
	 * @dataProvider dataGetUserAgent
	 *
	 * @param string $userAgent
	 */
	public function testGetUserAgent($userAgent) {
		$context = $this->getInstance();

		$this->request->expects($this->once())
			->method('getHeader')
			->with('User-Agent')
			->willReturn($userAgent);

		$this->assertSame($userAgent, $context->getUserAgent());
	}

	public function dataGetClientDevice() {
		return [
			['Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:40.0) Gecko/20100101 Firefox/40.0', Context::SYNC_CLIENT_OTHER],
			['Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/45.0.2454.85 Chrome/45.0.2454.85 Safari/537.36', Context::SYNC_CLIENT_OTHER],

			['mirall', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Windows) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Macintosh) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Linux) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (DragonFlyBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (FreeBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (NetBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (OpenBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Solaris) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Ubuntu) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Unknown OS) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Android) ownCloud-android/1.8.0', Context::SYNC_CLIENT_ANDROID],
			['Mozilla/5.0 (iOS) ownCloud-iOS/1.8.0', Context::SYNC_CLIENT_IOS],
			['ownCloudApp/1.3.1 (App/154; iOS/13.3; iPhone)', Context::SYNC_CLIENT_IOS],

			// Older user agents
			['csync', Context::SYNC_CLIENT_DESKTOP],

			['iOS-ownCloud', Context::SYNC_CLIENT_IOS],
			['ownCloud iOS Client', Context::SYNC_CLIENT_IOS],

			['Android-ownCloud', Context::SYNC_CLIENT_ANDROID],
			['Mozilla/5.0 (Android) ownCloud', Context::SYNC_CLIENT_ANDROID],

			// Branded clients
			['Special branded desktop client', Context::SYNC_CLIENT_DESKTOP_BRANDED],
			['Special branded android client', Context::SYNC_CLIENT_ANDROID_BRANDED],
			['Special branded ios client', Context::SYNC_CLIENT_IOS_BRANDED],
		];
	}

	/**
	 * @dataProvider dataGetClientDevice
	 *
	 * @param string $userAgent
	 * @param string $device
	 */
	public function testGetClientDevice($userAgent, $device) {
		$this->config->expects($this->once())
			->method('getBrandedClients')
			->willReturn([
				// Current user agents
				// Mozilla/5.0 (<os>) mirall/<version>
				'Special branded desktop client' => Context::SYNC_CLIENT_DESKTOP_BRANDED,
				'Special branded android client' => Context::SYNC_CLIENT_ANDROID_BRANDED,
				'Special branded ios client' => Context::SYNC_CLIENT_IOS_BRANDED,
			]);

		$context = $this->getInstance(['getUserAgent']);

		$context->expects($this->once())
			->method('getUserAgent')
			->willReturn($userAgent);

		$this->assertSame($device, $context->getClientDevice());
	}

	public function dataUserAgentMatchesDevice() {
		return [
			['special branded desktop client', '*b*', false],
			['special branded desktop client', 'branded', false],
			['special branded desktop client', '*branded', false],
			['special branded desktop client', '*Branded', false],
			['special branded desktop client', '*branded desktop client', true],
			['special branded desktop client', '*branded desktop Client', true],
			['special branded desktop client', 'branded*', false],
			['special branded desktop client', 'Branded*', false],
			['special branded desktop client', 'special branded*', true],
			['special branded desktop client', 'Special branded*', true],
			['special branded desktop client', '*branded*', true],
			['special branded desktop client', '*Branded*', true],
			['special branded desktop client', 'Special branded desktop client', true],
			['special branded desktop client', 'special branded desktop client', true],
			['special branded desktop client', 'special*desktop*', true],
			['special branded desktop client', 'special*client*', true],
			['special branded desktop client', 'special*other*', false],
			['special branded desktop client', 'branded*desktop*', false],

			['branded', 'special branded desktop client', false],
			['branded', '*special branded desktop client', false],
			['branded', 'special branded desktop client*', false],
			['branded', '*special branded desktop client*', false],
		];
	}

	/**
	 * @dataProvider dataUserAgentMatchesDevice
	 *
	 * @param string $userAgent
	 * @param string $device
	 * @param bool $expected
	 */
	public function testUserAgentMatchesDevice($userAgent, $device, $expected) {
		$context = $this->getInstance();
		$this->assertSame($expected, $this->invokePrivate($context, 'userAgentMatchesDevice', [$userAgent, $device]));
	}

	public function dataFileIds() {
		/**
		 * 1. inital set of ids that will be given to setFileIds
		 * 2. expected result array
		 * 3. folder structure from the Db. setFileIds will look into that
		 * and find out if one of the ids from 1. is a thumbnail and if yes
		 * find the source of that thumbnail and all the parents of the source
		 */
		return [
			[['10'], [10]],
			[['string'], []],
			[[42], [42]],
			[null, []],
			[[42, 21, 42, -1], [21, 42]],
			[[47, 22, 1], [1, 22, 32, 33, 34, 47],
				[
					47 => [
						"type" => \OCP\Files\FileInfo::TYPE_FILE,
						"name" => "sssdd.jpg",
					],
					22 => [
						"type" => \OCP\Files\FileInfo::TYPE_FOLDER,
						"name" => "34",
						"path" => "/admin/thumbnails/34",
					],
					34 => [
						"parent" => 33,
						"name" => "/admin/files",
					],
					33 => [
						"parent" => 32,
						"name" => "/admin",
					],
					32 => [
						"parent" => 1,
						"name" => "/"
					],
					1 => [
						"parent" => 1,
						"name" => "/"
					]
				]
			],
			[[55, 78, 32], [32, 55, 78],
				[
					55 => [
						"name" => "sssdd.jpg",
					],
					78 => [
						"name" => "xxxxx.jpg",
					],
					32 => [
						"type" => \OCP\Files\FileInfo::TYPE_FOLDER,
						"name" => "3456",
						"path" => "/admin/files/thumbnails/3456",
					]
				]
			],
			[[55, 78, 32], [32, 55, 78],
				[
					55 => [
						"name" => "sssdd.jpg",
					],
					78 => [
						"name" => "xxxxx.jpg",
					],
					32 => [
						"type" => \OCP\Files\FileInfo::TYPE_FILE,
						"name" => "3456",
						"path" => "/admin/thumbnails/3456",
					]
				]
			],
			[[55, 78, 32], [32, 55, 78],
				[
					55 => [
						"name" => "sssdd.jpg",
					],
					78 => [
						"name" => "xxxxx.jpg",
					],
					32 => [
						"type" => \OCP\Files\FileInfo::TYPE_FILE,
						"name" => "3456-45.jpg",
						"path" => "/admin/thumbnails/3456/45.jpg",
					]
				]
			],
			[[47, 22, 1], [1, 22, 33, 34, 47],
				[
					47 => [
						"type" => \OCP\Files\FileInfo::TYPE_FILE,
						"name" => "sssdd.jpg",
					],
					22 => [
						"type" => \OCP\Files\FileInfo::TYPE_FOLDER,
						"name" => "34",
						"path" => "/admin/thumbnails/34",
					],
					34 => [
						"parent" => 33,
						"name" => "/admin/files",
					],
					33 => [
						"parent" => 34,
						"name" => "/admin",
					],
					1 => [
						"parent" => 1,
						"name" => "/",
					],
				]
			],
		];
	}

	/**
	 * @dataProvider dataFileIds
	 *
	 * @param mixed $real
	 * @param array $expected
	 */
	public function testSetFileId($real, array $expected, array $filesInDb = null) {
		$nodesMap = [];
		if ($filesInDb !== null) {
			//mock general functions of the Nodes
			foreach ($filesInDb as $id => $file) {
				$node = $this->createMock('OCP\Files\Node');
				if (isset($file["type"])) {
					$node->method('getType')->willReturn($file["type"]);
				}
				if (isset($file["name"])) {
					$node->method('getName')->willReturn($file["name"]);
				}

				if (isset($file["path"])) {
					$node->method('getPath')->willReturn($file["path"]);
				}
				$node->method('getId')->willReturn($id);
				$parentFiles[$id] = $node;
			}
			//mock "getParent" by assigning the correct previously created mock
			//to the 'getParent' function
			foreach ($parentFiles as $id => $node) {
				if (isset($filesInDb [$id] ["parent"])) {
					$node->method('getParent')
						->willReturn($parentFiles [$filesInDb [$id] ["parent"]]);
				}
				//build the map for getById
				//when asked for $id it will return [$node]
				$nodesMap[] = [$id, false, [$node]];
			}
		} else {
			foreach ($expected as $id) {
				$node = $this->createMock('OCP\Files\Node');
				$nodesMap[] = [(int)$id, false, [$node]];
			}
		}

		$this->rootFolder->expects($this->any())
			->method('getById')
			->will($this->returnValueMap($nodesMap));

		$context = $this->getInstance();
		$this->assertSame($expected, $context->setFileIds($real));
		$this->assertSame($expected, $this->invokePrivate($context, 'fileIds'));
		$this->assertSame($expected, $context->getFileIds());
	}
}
