<?php
/**
 * ownCloud Firewall
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright Copyright (c) 2018, ownCloud GmbH
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Firewall\Tests\Lib;

use Test\TestCase;
use OCA\Firewall\Firewall;
use OCA\Firewall\StorageWrapper;
use OCP\Files\ForbiddenException;
use OCP\Files\IHomeStorage;
use OCP\Files\Mount\IMountManager;
use OCP\Files\Mount\IMountPoint;

class StorageWrapperTest extends TestCase {

	/** @var \OC\Files\Storage\Storage|\PHPUnit\Framework\MockObject\MockObject */
	protected $storage;

	/** @var \OCA\Firewall\Firewall|\PHPUnit\Framework\MockObject\MockObject */
	protected $firewall;

	/** @var IMountManager | \PHPUnit\Framework\MockObject\MockObject */
	protected $mountManager;

	protected function setUp(): void {
		parent::setUp();

		$this->storage = $this->createMock(\OCP\Files\Storage::class);
		$this->firewall = $this->createMock(Firewall::class);

		$this->mountManager = $this->createMock(IMountManager::class);
		$this->overwriteService('MountManager', $this->mountManager);
	}

	protected function tearDown(): void {
		$this->restoreService('MountManager');
	}

	/**
	 * @param array $methods
	 * @return StorageWrapper|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getInstance(array $methods = []) {
		if (empty($methods)) {
			return new StorageWrapper([
				'storage' => $this->storage,
				'mountPoint' => 'mountPoint',
				'firewall' => $this->firewall,
			]);
		} else {
			return $this->getMockBuilder(StorageWrapper::class)
				->setConstructorArgs([
					[
						'storage' => $this->storage,
						'mountPoint' => 'mountPoint',
						'firewall' => $this->firewall,
					]
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataCheckFileInHome() {
		return [
			[
				'files/path1/folder/file',
				[
					['files/path1/folder/file', '1'],
					['files/path1/folder', '-1'],
					['files/path1', '-1'],
					['files', '-1'],
					['', '-1'],
				],
				[1],
				true,
			],
			[
				'thumbnails/path1/folder/file',
				[
					['thumbnails/path1/folder/file', '1'],
					['thumbnails/path1/folder', '-1'],
					['thumbnails/path1', '-1'],
					['thumbnails', '-1'],
					['', '-1'],
				],
				[1],
				true,
			],
			[
				'files/path2/folder/file',
				[
					['files/path2/folder/file', '42'],
					['files/path2/folder', '21'],
					['files/path2', '32'],
					['files', '16'],
					['', '8'],
				],
				[42, 21, 32, 16, 8],
				false,
			],
			[
				'files',
				[
					['files', '16'],
					['', '8'],
				],
				[16, 8],
				false,
			],
		];
	}

	/**
	 * Tests that the firewall check applies to the home storage
	 *
	 * @dataProvider dataCheckFileInHome
	 * @param string $path path to check
	 * @param array $cacheReturns cache return value
	 * @param array $expectedIds expected ids
	 * @param bool $throws whether an expected exception is thrown
	 */
	public function testCheckFileInHome($path, array $cacheReturns, array $expectedIds, $throws) {
		$wrapper = $this->getInstance();

		$this->storage->expects($this->once())
			->method('instanceOfStorage')
			->with(IHomeStorage::class)
			->willReturn(true);

		$cache = $this->createMock(\OC\Files\Cache\Cache::class);
		$cache->expects($this->exactly(\sizeof($cacheReturns)))
			->method('getId')
			->willReturnMap($cacheReturns);

		$this->storage->expects($this->any())
			->method('getCache')
			->willReturn($cache);

		if ($throws) {
			$this->firewall->expects($this->once())
				->method('checkRulesForFiles')
				->with($expectedIds)
				->willThrowException(new ForbiddenException('', true));

			$this->expectException(ForbiddenException::class);
		} else {
			$this->firewall->expects($this->once())
				->method('checkRulesForFiles')
				->with($expectedIds);
		}

		$this->invokePrivate($wrapper, 'checkFirewall', [$path]);
	}

	public function dataCheckFileInExternalStorage() {
		return [
			[
				'path1/folder/file',
				[
					['path1/folder/file', '1'],
					['path1/folder', '-1'],
					['path1', '-1'],
					['', '-1'],
				],
				[1],
				true,
			],
			[
				'path2/folder/file',
				[
					['path2/folder/file', '42'],
					['path2/folder', '21'],
					['path2', '32'],
					['', '16'],
				],
				[42, 21, 32, 16],
				false,
			],
			[
				'files',
				[
					['files', '16'],
					['', '8'],
				],
				[16, 8],
				false,
			],
			[
				'thumbnails',
				[
					['thumbnails', '16'],
					['', '8'],
				],
				[16, 8],
				false,
			],
		];
	}

	/**
	 * Tests that the firewall check applies to an external storage
	 *
	 * @dataProvider dataCheckFileInExternalStorage
	 * @param string $path path to check
	 * @param array $cacheReturns cache return value
	 * @param array $expectedIds expected ids
	 * @param bool $throws whether an expected exception is thrown
	 */
	public function testCheckFileInExternalStorage($path, array $cacheReturns, array $expectedIds, $throws) {
		$wrapper = $this->getInstance();

		$this->storage->expects($this->once())
			->method('instanceOfStorage')
			->with(IHomeStorage::class)
			->willReturn(false);

		$mount = $this->createMock(IMountPoint::class);
		$mount->expects($this->once())
			->method('getMountPoint')
			->willReturn('/user1/files/smb');

		$this->mountManager->expects($this->once())
			->method('findByStorageId')
			->with($wrapper->getId())
			->willReturn([$mount]);

		$cache = $this->createMock(\OC\Files\Cache\Cache::class);
		$cache->expects($this->exactly(\sizeof($cacheReturns)))
			->method('getId')
			->willReturnMap($cacheReturns);

		$this->storage->expects($this->any())
			->method('getCache')
			->willReturn($cache);

		if ($throws) {
			$this->firewall->expects($this->once())
				->method('checkRulesForFiles')
				->with($expectedIds)
				->willThrowException(new ForbiddenException('', true));

			$this->expectException(ForbiddenException::class);
		} else {
			$this->firewall->expects($this->once())
				->method('checkRulesForFiles')
				->with($expectedIds);
		}

		$this->invokePrivate($wrapper, 'checkFirewall', [$path]);
	}

	/**
	 * Tests that the firewall checks are skipped whenever the path is
	 * in the home storage but outside of "files/"
	 */
	public function testDoesNotCheckFileInHomeOutsideFiles() {
		$wrapper = $this->getInstance();

		$this->storage->expects($this->once())
			->method('instanceOfStorage')
			->with(IHomeStorage::class)
			->willReturn(true);

		$this->firewall->expects($this->never())
			->method('checkRulesForFiles');

		$this->invokePrivate($wrapper, 'checkFirewall', ['files_encryption/somekey.key']);
	}

	/**
	 * Tests that the firewall checks are skipped whenever the path is
	 * in the root storage.
	 */
	public function testDoesNotCheckFileRootStorage() {
		$wrapper = $this->getInstance();

		$mount = $this->createMock(IMountPoint::class);
		$mount->expects($this->once())
			->method('getMountPoint')
			->willReturn('/');

		$this->mountManager->expects($this->once())
			->method('findByStorageId')
			->with($wrapper->getId())
			->willReturn([$mount]);

		$this->storage->expects($this->once())
			->method('instanceOfStorage')
			->with(IHomeStorage::class)
			->willReturn(false);

		$this->firewall->expects($this->never())
			->method('checkRulesForFiles');

		$this->invokePrivate($wrapper, 'checkFirewall', ['files_encryption/somekey.key']);
	}

	/*
	 * Storage Wrapper methods
	 * =======================================================================
	 */

	public function dataOnePath() {
		return [
			['path1', true],
			['path2', false],
		];
	}

	public function dataTwoStrings() {
		return [
			['path1', 'data1', true],
			['path2', 'data2', false],
		];
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testMkdir($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('mkdir')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->mkdir($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testRmdir($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('rmdir')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->rmdir($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testIsCreatable($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('isCreatable')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->isCreatable($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testIsReadable($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('isReadable')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->isReadable($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testIsUpdatable($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('isUpdatable')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->isUpdatable($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testIsDeletable($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('isDeletable')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->isDeletable($path));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testFileGetContents($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('file_get_contents')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->file_get_contents($path));
	}

	/**
	 * @dataProvider dataTwoStrings
	 * @param string $path
	 * @param string $data
	 * @param bool $expected
	 */
	public function testFilePutContents($path, $data, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('file_put_contents')
			->with($path, $data)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->file_put_contents($path, $data));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testUnlink($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('unlink')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->unlink($path));
	}

	/**
	 * @dataProvider dataTwoStrings
	 * @param string $path1
	 * @param string $path2
	 * @param bool $expected
	 */
	public function testRename($path1, $path2, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('rename')
			->with($path1, $path2)
			->willReturn($expected);

		$wrapper->expects($this->at(0))
			->method('checkFirewall')
			->with($path1);

		$wrapper->expects($this->at(1))
			->method('checkFirewall')
			->with($path2);

		$this->assertSame($expected, $wrapper->rename($path1, $path2));
	}

	/**
	 * @dataProvider dataTwoStrings
	 * @param string $path1
	 * @param string $path2
	 * @param bool $expected
	 */
	public function testCopy($path1, $path2, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('copy')
			->with($path1, $path2)
			->willReturn($expected);

		$wrapper->expects($this->at(0))
			->method('checkFirewall')
			->with($path1);

		$wrapper->expects($this->at(1))
			->method('checkFirewall')
			->with($path2);

		$this->assertSame($expected, $wrapper->copy($path1, $path2));
	}

	/**
	 * @dataProvider dataTwoStrings
	 * @param string $path
	 * @param string $mode
	 * @param bool $expected
	 */
	public function testFopen($path, $mode, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('fopen')
			->with($path, $mode)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->fopen($path, $mode));
	}

	public function dataTouch() {
		return [
			['path1', 12, true],
			['path2', 32, false],
		];
	}

	/**
	 * @dataProvider dataTouch
	 * @param string $path
	 * @param int $time
	 * @param bool $expected
	 */
	public function testTouch($path, $time, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('touch')
			->with($path, $time)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->touch($path, $time));
	}

	/**
	 * @dataProvider dataOnePath
	 * @param string $path
	 * @param bool $expected
	 */
	public function testGetDirectDownload($path, $expected) {
		$wrapper = $this->getInstance(['checkFirewall']);

		$this->storage->expects($this->once())
			->method('getDirectDownload')
			->with($path)
			->willReturn($expected);

		$wrapper->expects($this->once())
			->method('checkFirewall')
			->with($path);

		$this->assertSame($expected, $wrapper->getDirectDownload($path));
	}

	public function dataXFromStorage() {
		return [
			[true, 'source1', 'target1', true],
			[true, 'source2', 'target2', false],
			[false, 'source1', 'target1', true],
			[false, 'source2', 'target2', false],
		];
	}

	/**
	 * @dataProvider dataXFromStorage
	 * @param bool $selfStorage
	 * @param string $sourcePath
	 * @param string $targetPath
	 * @param bool $expected
	 */
	public function testCopyFromStorage($selfStorage, $sourcePath, $targetPath, $expected) {
		$wrapper = $this->getInstance([
			'checkFirewall',
			'copy',
		]);

		if ($selfStorage) {
			$sourceStorage = $wrapper;

			$wrapper->expects($this->once())
				->method('copy')
				->with($sourcePath, $targetPath)
				->willReturn($expected);
		} else {
			$sourceStorage = $this->createMock(\OCP\Files\Storage::class);

			$wrapper->expects($this->once())
				->method('checkFirewall')
				->with($targetPath);

			$this->storage->expects($this->once())
				->method('copyFromStorage')
				->with($sourceStorage, $sourcePath, $targetPath)
				->willReturn($expected);
		}

		$this->assertSame($expected, $wrapper->copyFromStorage($sourceStorage, $sourcePath, $targetPath));
	}

	/**
	 * @dataProvider dataXFromStorage
	 * @param bool $selfStorage
	 * @param string $sourcePath
	 * @param string $targetPath
	 * @param bool $expected
	 */
	public function testMoveFromStorage($selfStorage, $sourcePath, $targetPath, $expected) {
		$wrapper = $this->getInstance([
			'checkFirewall',
			'rename',
		]);

		if ($selfStorage) {
			$sourceStorage = $wrapper;

			$wrapper->expects($this->once())
				->method('rename')
				->with($sourcePath, $targetPath)
				->willReturn($expected);
		} else {
			$sourceStorage = $this->createMock(\OCP\Files\Storage::class);

			$wrapper->expects($this->once())
				->method('checkFirewall')
				->with($targetPath);

			$this->storage->expects($this->once())
				->method('moveFromStorage')
				->with($sourceStorage, $sourcePath, $targetPath)
				->willReturn($expected);
		}

		$this->assertSame($expected, $wrapper->moveFromStorage($sourceStorage, $sourcePath, $targetPath));
	}
}
