<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Condition;

use OCA\Workflow\Condition\Context;
use OCP\Files\IMimeTypeDetector;
use OCP\IConfig;
use OCP\IGroupManager;
use OCP\IRequest;
use OCP\IUser;
use OCP\IUserSession;
use Test\TestCase;

class ContextTest extends TestCase {

	/** @var IGroupManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $groupManager;

	/** @var IRequest|\PHPUnit\Framework\MockObject\MockObject */
	protected $request;

	/** @var IUserSession|\PHPUnit\Framework\MockObject\MockObject */
	protected $userSession;

	/** @var IMimeTypeDetector|\PHPUnit\Framework\MockObject\MockObject */
	protected $mimeTypeDetector;

	/** @var IConfig|\PHPUnit\Framework\MockObject\MockObject */
	protected $config;

	protected function setUp(): void {
		parent::setUp();

		$this->groupManager = $this->getMockBuilder(IGroupManager::class)
			->disableOriginalConstructor()
			->getMock();

		$this->request = $this->getMockBuilder(IRequest::class)
			->disableOriginalConstructor()
			->getMock();

		$this->userSession = $this->getMockBuilder(IUserSession::class)
			->disableOriginalConstructor()
			->getMock();

		$this->mimeTypeDetector = $this->getMockBuilder(IMimeTypeDetector::class)
			->disableOriginalConstructor()
			->getMock();

		$this->config = $this->getMockBuilder(IConfig::class)
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return Context|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getInstance($methods = []) {
		if (empty($methods)) {
			return new Context(
				$this->request,
				$this->groupManager,
				$this->userSession,
				$this->mimeTypeDetector,
				$this->config
			);
		} else {
			return $this->getMockBuilder(\OCA\Workflow\Condition\Context::class)
				->setConstructorArgs([
					$this->request,
					$this->groupManager,
					$this->userSession,
					$this->mimeTypeDetector,
					$this->config,
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataGetRequestFullUrl() {
		return [
			['http', 'localhost', '/', 'http://localhost/'],
			['https', 'owncloud.org', '/index.php', 'https://owncloud.org/index.php'],
			['https', 'localhost', '/owncloud/index.php', 'https://localhost/owncloud/index.php'],
		];
	}

	/**
	 * @dataProvider dataGetRequestFullUrl
	 *
	 * @param string $protocol
	 * @param string $host
	 * @param string $requestUri
	 * @param string $expected
	 */
	public function testGetRequestFullUrl($protocol, $host, $requestUri, $expected) {
		$context = $this->getInstance();

		$this->request->expects($this->any())
			->method('getServerProtocol')
			->willReturn($protocol);

		$this->request->expects($this->any())
			->method('getServerHost')
			->willReturn($host);

		$this->request->expects($this->any())
			->method('getRequestUri')
			->willReturn($requestUri);

		$this->assertSame($expected, $context->getRequestFullUrl());
	}

	public function dataGetRequestUri() {
		return [
			['/'],
			['/index.php'],
			['/owncloud/index.php'],
		];
	}

	/**
	 * @dataProvider dataGetRequestUri
	 *
	 * @param string $requestUri
	 */
	public function testGetRequestUri($requestUri) {
		$context = $this->getInstance();

		$this->request->expects($this->any())
			->method('getRequestUri')
			->willReturn($requestUri);

		$this->assertSame($requestUri, $context->getRequestUri());
	}

	public function dataGetRequestType() {
		return [
			['/index.php', '/apps/files', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/index.php', '/apps/files', true, 'user1', Context::REQUEST_TYPE_OTHER],
			['/public.php', '/webdav', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '/webdav', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '?service=files&token=tokenis13char', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/public.php', '?service=files&token=tokenis13char', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/index.php', '/s/tokenis13char', false, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/index.php', '/s/tokenis13char', true, 'user1', Context::REQUEST_TYPE_PUBLIC],
			['/remote.php', '/webdav', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/some/path', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/webdav/some/path', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/user1/path', false, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/files/user1/path', true, 'user1', Context::REQUEST_TYPE_WEBDAV],
			['/remote.php', '/dav/addressbooks', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks', true, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks/', false, 'user1', Context::REQUEST_TYPE_OTHER],
			['/remote.php', '/dav/addressbooks/', true, 'user1', Context::REQUEST_TYPE_OTHER],

		];
	}

	/**
	 * @dataProvider dataGetRequestType
	 *
	 * @param string $scriptName
	 * @param string $pathInfo
	 * @param bool $verbose
	 * @param string|null $sessionUser
	 * @param string $expected
	 */
	public function testGetRequestType($scriptName, $pathInfo, $verbose, $sessionUser, $expected) {
		$context = $this->getInstance();

		if ($sessionUser === null) {
			$this->userSession->expects($this->any())
				->method('getUser')
				->willReturn(null);
		} else {
			$this->userSession->expects($this->any())
				->method('getUser')
				->willReturn($this->createMock(IUser::class));
		}

		$this->request->expects($this->atLeastOnce())
			->method('getScriptName')
			->willReturn($scriptName);

		$this->request->expects($this->any())
			->method('getPathInfo')
			->willReturn($pathInfo);

		$this->assertSame($expected, $context->getRequestType($verbose));
	}

	public function testGetRequestTime() {
		$context = $this->getInstance();

		$before = \time();
		$actual = $context->getRequestTime();
		$after = \time();

		$this->assertGreaterThanOrEqual($before, $actual);
		$this->assertLessThanOrEqual($after, $actual);
	}

	public function dataGetUploadSize() {
		return [
			[2, null, null, 0],
			[1, 0, null, 0],
			[1, 1, null, 1],
			[1, 0, 1, 0],
			[1, 1, 0, 1],
			[2, null, 0, 0],
			[2, null, 1, 1],
		];
	}

	/**
	 * @dataProvider dataGetUploadSize
	 *
	 * @param int $headerCalls
	 * @param int $ocTotalLength
	 * @param int $contentLength
	 * @param int $expected
	 */
	public function testGetUploadSizeSizeHandling($headerCalls, $ocTotalLength, $contentLength, $expected) {
		$context = $this->getInstance([
			'getRequestType'
		]);

		$context->expects($this->any())
			->method('getRequestType')
			->willReturn(Context::REQUEST_TYPE_WEBDAV);

		$this->request->expects($this->any())
			->method('getMethod')
			->willReturn('PUT');

		$this->request->expects($this->exactly($headerCalls))
			->method('getHeader')
			->willReturnMap([
				['OC_TOTAL_LENGTH', $ocTotalLength],
				['CONTENT_LENGTH', $contentLength],
			]);

		$this->assertSame($expected, $context->getUploadSize());
	}

	public function dataGetUploadSizeCheck() {
		return [
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', null, 10],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', null, 0],

			[Context::REQUEST_TYPE_PUBLIC, 'PUT', null, 0],
			[Context::REQUEST_TYPE_PUBLIC, 'PUT', ['not_empty'], 10],

			[Context::REQUEST_TYPE_OTHER, 'GET', null, 0],
			[Context::REQUEST_TYPE_OTHER, 'POST', [], 0],
			[Context::REQUEST_TYPE_OTHER, 'POST', ['not_empty'], 0],
			[Context::REQUEST_TYPE_OTHER, 'PUT', [], 0],
			[Context::REQUEST_TYPE_OTHER, 'PUT', ['not_empty'], 10],
		];
	}

	/**
	 * @dataProvider dataGetUploadSizeCheck
	 *
	 * @param string $requestType
	 * @param string $requestMethod
	 * @param array|null $uploadedFiles
	 * @param bool $expected
	 */
	public function testGetUploadSizeCheck($requestType, $requestMethod, $uploadedFiles, $expected) {
		$context = $this->getInstance([
			'getRequestType'
		]);

		if ($requestType === Context::REQUEST_TYPE_WEBDAV) {
			$context->method('getRequestType')
				->willReturn($requestType);

			$this->request->method('getMethod')
				->willReturn($requestMethod);
		} else {
			$context->method('getRequestType')
				->willReturn($requestType);

			$this->request->method('getMethod')
				->willReturn($requestMethod);

			$this->request->method('getUploadedFile')
				->willReturn($uploadedFiles);
		}

		$this->request->method('getHeader')
			->willReturn($expected);

		$this->assertSame($expected, $context->getUploadSize());
	}

	public function dataGetUploadType() {
		return [
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', 'test.txt', [], 'text/plain'],
			[Context::REQUEST_TYPE_WEBDAV, 'PUT', 'test.mp3', [], 'audio/mpeg'],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', '', [], null],
			[Context::REQUEST_TYPE_WEBDAV, 'PROPFIND', 'test.mp3', [], null],

			[Context::REQUEST_TYPE_OTHER, 'GET', '', [], null],
			[Context::REQUEST_TYPE_OTHER, 'GET', '', [
				'type' => [
				],
			], null],
			[Context::REQUEST_TYPE_OTHER, 'GET', '', [
				'type' => [
					0 => 'text/plain'
				],
			], null],

			[Context::REQUEST_TYPE_OTHER, 'POST', '', [], null],
			[Context::REQUEST_TYPE_OTHER, 'POST', '', [
				'type' => [
				],
			], null],
			[Context::REQUEST_TYPE_OTHER, 'POST', '', [
				'type' => [
					0 => 'text/plain'
				],
			], 'text/plain'],
			[Context::REQUEST_TYPE_OTHER, 'PUT', '', [
				'type' => [
					0 => 'text/plain'
				],
			], 'text/plain'],

			[Context::REQUEST_TYPE_PUBLIC, 'POST', '', [], null],
			[Context::REQUEST_TYPE_PUBLIC, 'POST', '', [
				'type' => null,
			], null],
			[Context::REQUEST_TYPE_PUBLIC, 'PUT', '', [], null],
		];
	}

	/**
	 * @dataProvider dataGetUploadType
	 *
	 * @param string $requestType
	 * @param string $requestMethod
	 * @param string $webDavPath
	 * @param array $files
	 * @param mixed $expected
	 */
	public function testGetUploadType($requestType, $requestMethod, $webDavPath, $files, $expected) {
		$context = $this->getInstance([
			'getRequestType'
		]);

		if ($requestType === Context::REQUEST_TYPE_WEBDAV) {
			$context->method('getRequestType')
				->willReturn($requestType);

			$this->request->method('getMethod')
				->willReturn($requestMethod);

			$this->request->method('getPathInfo')
				->willReturn($webDavPath);

			$this->mimeTypeDetector->method('detectPath')
				->willReturnMap([
					['test.txt', 'text/plain'],
					['test.mp3', 'audio/mpeg'],
				]);
		} else {
			$context->method('getRequestType')
				->willReturn($requestType);

			$this->request->expects($this->any())
				->method('getMethod')
				->willReturn($requestMethod);

			$this->request->method('getUploadedFile')
				->with('files')
				->willReturn($files);
		}

		$this->assertSame($expected, $context->getUploadType());
	}

	public function dataGetUserGroups() {
		$user = $this->getMockBuilder('OCP\IUser')
			->disableOriginalConstructor()
			->getMock();

		return [
			[null, []],
			[$user, []],
			[$user, ['admin']],
			[$user, ['admin', 'test1']],
		];
	}

	/**
	 * @dataProvider dataGetUserGroups
	 *
	 * @param mixed $user
	 * @param array $expected
	 */
	public function testGetUserGroups($user, array $expected) {
		$context = $this->getInstance();

		$this->userSession->expects($this->any())
			->method('getUser')
			->willReturn($user);

		if ($user !== null) {
			$this->groupManager->expects($this->any())
				->method('getUserGroupIds')
				->with($user)
				->willReturn($expected);
		} else {
			$this->groupManager->expects($this->never())
				->method('getUserGroupIds');
		}

		$this->assertSame($expected, $context->getUserGroups());
	}

	public function dataGetRemoteAddress() {
		return [
			['127.0.0.1'],
			['192.168.2.100'],
		];
	}

	/**
	 * @dataProvider dataGetRemoteAddress
	 *
	 * @param string $remoteAddress
	 */
	public function testGetRemoteAddress($remoteAddress) {
		$context = $this->getInstance();

		$this->request->expects($this->any())
			->method('getRemoteAddress')
			->willReturn($remoteAddress);

		$this->assertSame($remoteAddress, $context->getRemoteAddress());
	}

	public function dataGetServerAddress() {
		return [
			['127.0.0.1', '127.0.0.1'],
			['', ''],
			[null, ''],
		];
	}

	/**
	 * @dataProvider dataGetServerAddress
	 *
	 * @param string $remoteServerAddress
	 * @param string $expected
	 */
	public function testGetServerAddress($remoteServerAddress, $expected) {
		$context = $this->getInstance();

		$this->request->expects($this->any())
			->method('getHeader')
			->with('SERVER_ADDR')
			->willReturn($remoteServerAddress);

		$this->assertSame($expected, $context->getServerAddress());
	}

	public function dataGetUserAgent() {
		return [
			['Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:40.0) Gecko/20100101 Firefox/40.0'],
			['Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/45.0.2454.85 Chrome/45.0.2454.85 Safari/537.36'],
		];
	}

	/**
	 * @dataProvider dataGetUserAgent
	 *
	 * @param string $userAgent
	 */
	public function testGetUserAgent($userAgent) {
		$context = $this->getInstance();

		$this->request->expects($this->any())
			->method('getHeader')
			->with('User-Agent')
			->willReturn($userAgent);

		$this->assertSame($userAgent, $context->getUserAgent());
	}

	public function dataGetClientDevice() {
		return [
			['Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:40.0) Gecko/20100101 Firefox/40.0', Context::SYNC_CLIENT_OTHER],
			['Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/45.0.2454.85 Chrome/45.0.2454.85 Safari/537.36', Context::SYNC_CLIENT_OTHER],

			['mirall', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Windows) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Macintosh) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Linux) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (DragonFlyBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (FreeBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (NetBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (OpenBSD) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Solaris) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Ubuntu) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Unknown OS) mirall/2.0.2', Context::SYNC_CLIENT_DESKTOP],
			['Mozilla/5.0 (Android) ownCloud-android/1.8.0', Context::SYNC_CLIENT_ANDROID],
			['Mozilla/5.0 (iOS) ownCloud-iOS/1.8.0', Context::SYNC_CLIENT_IOS],

			// Older user agents
			['csync', Context::SYNC_CLIENT_DESKTOP],

			['iOS-ownCloud', Context::SYNC_CLIENT_IOS],
			['ownCloud iOS Client', Context::SYNC_CLIENT_IOS],

			['Android-ownCloud', Context::SYNC_CLIENT_ANDROID],
			['Mozilla/5.0 (Android) ownCloud', Context::SYNC_CLIENT_ANDROID],

			// Branded clients
			['Special branded desktop client', Context::SYNC_CLIENT_DESKTOP_BRANDED],
			['Special branded android client', Context::SYNC_CLIENT_ANDROID_BRANDED],
			['Special branded ios client', Context::SYNC_CLIENT_IOS_BRANDED],
		];
	}

	/**
	 * @dataProvider dataGetClientDevice
	 *
	 * @param string $userAgent
	 * @param string $device
	 */
	public function testGetClientDevice($userAgent, $device) {
		$context = $this->getInstance([
			'getUserAgent',
			'getBrandedClients'
		]);

		$context->expects($this->any())
			->method('getBrandedClients')
			->willReturn([
				// Current user agents
				// Mozilla/5.0 (<os>) mirall/<version>
				'Special branded desktop client' => Context::SYNC_CLIENT_DESKTOP_BRANDED,
				'Special branded android client' => Context::SYNC_CLIENT_ANDROID_BRANDED,
				'Special branded ios client' => Context::SYNC_CLIENT_IOS_BRANDED,
			]);

		$context->expects($this->any())
			->method('getUserAgent')
			->willReturn($userAgent);

		$this->assertSame($device, $context->getClientDevice());
	}

	public function dataUserAgentMatchesDevice() {
		return [
			['special branded desktop client', '*b*', false],
			['special branded desktop client', 'branded', false],
			['special branded desktop client', '*branded', false],
			['special branded desktop client', '*Branded', false],
			['special branded desktop client', '*branded desktop client', true],
			['special branded desktop client', '*branded desktop Client', true],
			['special branded desktop client', 'branded*', false],
			['special branded desktop client', 'Branded*', false],
			['special branded desktop client', 'special branded*', true],
			['special branded desktop client', 'Special branded*', true],
			['special branded desktop client', '*branded*', true],
			['special branded desktop client', '*Branded*', true],
			['special branded desktop client', 'Special branded desktop client', true],
			['special branded desktop client', 'special branded desktop client', true],

			['branded', 'special branded desktop client', false],
			['branded', '*special branded desktop client', false],
			['branded', 'special branded desktop client*', false],
			['branded', '*special branded desktop client*', false],
		];
	}

	/**
	 * @dataProvider dataUserAgentMatchesDevice
	 *
	 * @param string $userAgent
	 * @param string $device
	 * @param bool $expected
	 */
	public function testUserAgentMatchesDevice($userAgent, $device, $expected) {
		$context = $this->getInstance();
		$this->assertSame($expected, $this->invokePrivate($context, 'userAgentMatchesDevice', [$userAgent, $device]));
	}

	public function dataGetBrandedClients() {
		return [
			[false, []],
			[true, []],
			[null, []],
			[[], []],
			[['clients'], ['clients']],
		];
	}

	/**
	 * @dataProvider dataGetBrandedClients
	 *
	 * @param mixed $return
	 * @param array $expected
	 */
	public function testGetBrandedClients($return, $expected) {
		$context = $this->getInstance();

		$this->config->expects($this->any())
			->method('getSystemValue')
			->with('workflow.branded_clients')
			->willReturn($return);

		$this->assertSame($expected, $this->invokePrivate($context, 'getBrandedClients'));
	}

	public function dataFileIds() {
		return [
			[['10'], [10]],
			[['string'], []],
			[[42], [42]],
			[null, []],
			[[42, 21, 42, -1], [21, 42]],
		];
	}

	/**
	 * @dataProvider dataFileIds
	 *
	 * @param mixed $real
	 * @param array $expected
	 */
	public function testSetFileId($real, array $expected) {
		$context = $this->getInstance();
		$this->assertSame($expected, $context->setFileIds($real));
		$this->assertSame($expected, $this->invokePrivate($context, 'fileIds'));
		$this->assertSame($expected, $context->getFileIds());
	}
}
