<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Controller;

use OCA\Workflow\Controller\FlowController;
use OCA\Workflow\Engine\Exception\FlowDoesNotExist;
use OCA\Workflow\Engine\Flow;
use OCA\Workflow\Tests\Unit\TestCase;
use OCP\AppFramework\Http;

class FlowControllerTest extends TestCase {

	/** @var \OCA\Workflow\Engine\FlowManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $manager;

	/** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface|\PHPUnit\Framework\MockObject\MockObject */
	protected $dispatcher;

	/** @var \OCA\Workflow\Condition\RuleFactory|\PHPUnit\Framework\MockObject\MockObject */
	protected $ruleFactory;

	public function setUp(): void {
		parent::setUp();

		$this->manager = $this->getMockBuilder('OCA\Workflow\Engine\FlowManager')
			->disableOriginalConstructor()
			->getMock();

		$this->dispatcher = $this->getMockBuilder('Symfony\Component\EventDispatcher\EventDispatcherInterface')
			->disableOriginalConstructor()
			->getMock();

		$this->ruleFactory = $this->getMockBuilder('OCA\Workflow\Condition\RuleFactory')
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return FlowController|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getController(array $methods = []) {
		/** @var $request \OCP\IRequest|\PHPUnit\Framework\MockObject\MockObject */
		$request = $this->getMockBuilder('OCP\IRequest')
			->disableOriginalConstructor()
			->getMock();

		/** @var $l \OCP\IL10N|\PHPUnit\Framework\MockObject\MockObject */
		$l = $this->getMockBuilder('OCP\IL10N')
			->disableOriginalConstructor()
			->getMock();
		$l->expects($this->any())
			->method('t')
			->willReturnCallback(function ($string, $args = []) {
				return \vsprintf($string, $args);
			});

		if (empty($methods)) {
			return new FlowController(
				'workflow',
				$request,
				$this->manager,
				$this->dispatcher,
				$this->ruleFactory,
				$l
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\Controller\FlowController')
				->setConstructorArgs([
					'workflow',
					$request,
					$this->manager,
					$this->dispatcher,
					$this->ruleFactory,
					$l
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataGetWorkFlows() {
		return [
			[[], []],
			[[new Flow(1)], ['Flow#1']],
			[[new Flow(1), new Flow(2)], ['Flow#1', 'Flow#2']],
		];
	}

	/**
	 * @dataProvider dataGetWorkFlows
	 * @param Flow[] $flows
	 * @param array $expected
	 */
	public function testGetWorkFlows($flows, $expected) {
		$this->manager->expects($this->once())
			->method('getFlowsForTrigger')
			->with(null)
			->willReturn($flows);
		$this->manager->expects($this->exactly(\sizeof($flows)))
			->method('flowToArray')
			->willReturnCallback(function (Flow $flow) {
				return 'Flow#' . $flow->getId();
			});

		$controller = $this->getController();
		$response = $controller->getWorkFlows();

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals(Http::STATUS_OK, $response->getStatus());
		$this->assertEquals($expected, $response->getData());
	}

	public function dataAddWorkFlow() {
		return [
			['type1', 'name1', null, [], null, 1, Http::STATUS_OK, ['flowData' => 'Flow#', 'id' => 1]],
			['type1', 'name1', [], [], null, 1, Http::STATUS_OK, ['flowData' => 'Flow#', 'id' => 1]],
			['type2', 'name2', ['not empty'], [], null, 2, Http::STATUS_OK, ['flowData' => 'Flow#', 'id' => 2]],
			['type1', 'name1', [], ['not empty'], new \OutOfBoundsException('validation error'), null, Http::STATUS_BAD_REQUEST, ['error' => 'validation error']],
		];
	}

	/**
	 * @dataProvider dataAddWorkFlow
	 * @param string $type
	 * @param string $name
	 * @param array $conditions
	 * @param array $actions
	 * @param mixed $validationThrows
	 * @param mixed $managerReturns
	 * @param int $expectedStatus
	 * @param array $expectedData
	 */
	public function testAddWorkFlow($type, $name, $conditions, array $actions, $validationThrows, $managerReturns, $expectedStatus, $expectedData) {
		$controller = $this->getController(['validateWorkFlow']);

		if (!\is_array($conditions)) {
			$setConditions = [];
		} else {
			$setConditions = $conditions;
		}

		$flow = $this->getMockBuilder('OCA\Workflow\Engine\Flow')
			->disableOriginalConstructor()
			->getMock();
		$flow->expects($this->once())
			->method('setType')
			->with($type)
			->willReturnSelf();
		$flow->expects($this->once())
			->method('setName')
			->with($name)
			->willReturnSelf();
		$flow->expects($this->once())
			->method('setConditions')
			->with($setConditions)
			->willReturnSelf();
		$flow->expects($this->once())
			->method('setActions')
			->with($actions)
			->willReturnSelf();

		$this->manager->expects($this->once())
			->method('createFlow')
			->willReturn($flow);

		if ($validationThrows instanceof \Exception) {
			$controller->expects($this->once())
				->method('validateWorkFlow')
				->with($flow)
				->willThrowException($validationThrows);

			$this->manager->expects($this->never())
				->method('addFlow');
			$this->manager->expects($this->never())
				->method('flowToArray');
		} else {
			$controller->expects($this->once())
				->method('validateWorkFlow')
				->with($flow);

			$this->manager->expects($this->once())
				->method('addFlow')
				->with($flow)
				->willReturn($managerReturns);
			$this->manager->expects($this->once())
				->method('flowToArray')
				->with($flow)
				->willReturnCallback(function (Flow $flow) {
					return ['flowData' => 'Flow#' . $flow->getId()];
				});
		}

		$response = $controller->addWorkFlow($type, $name, $conditions, $actions);

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals($expectedStatus, $response->getStatus());
		$this->assertEquals($expectedData, $response->getData());
	}

	public function dataUpdateWorkFlow() {
		return [
			[1, 'rule1', null, [], null, true, Http::STATUS_OK, ['flowData' => 'Flow#1']],
			[1, 'rule1', [], [], null, true, Http::STATUS_OK, ['flowData' => 'Flow#1']],
			[2, 'rule2', ['not empty'], [], null, true, Http::STATUS_OK, ['flowData' => 'Flow#2']],
			[3, 'rule3', [], ['not empty'], new \OutOfBoundsException('validation error'), null, Http::STATUS_BAD_REQUEST, ['error' => 'validation error']],
			[4, 'rule4', [], [], null, new FlowDoesNotExist(), Http::STATUS_BAD_REQUEST, ['error' => 'The workflow does not exist']],
		];
	}

	/**
	 * @dataProvider dataUpdateWorkFlow
	 * @param int $ruleId
	 * @param string $name
	 * @param array $conditions
	 * @param array $actions
	 * @param mixed $validationThrows
	 * @param mixed $managerReturns
	 * @param int $expectedStatus
	 * @param array $expectedData
	 */
	public function testUpdateWorkFlow($ruleId, $name, $conditions, array $actions, $validationThrows, $managerReturns, $expectedStatus, $expectedData) {
		$controller = $this->getController(['validateWorkFlow']);

		if ($managerReturns instanceof \Exception) {
			$this->manager->expects($this->once())
				->method('getFlowById')
				->with($ruleId)
				->willThrowException($managerReturns);

			$controller->expects($this->never())
				->method('validateWorkFlow');
			$this->manager->expects($this->never())
				->method('updateFlow');
			$this->manager->expects($this->never())
				->method('flowToArray');
		} else {
			if (!\is_array($conditions)) {
				$setConditions = [];
			} else {
				$setConditions = $conditions;
			}

			$flow = $this->getMockBuilder('OCA\Workflow\Engine\Flow')
				->disableOriginalConstructor()
				->getMock();
			$flow->expects($this->once())
				->method('setName')
				->with($name)
				->willReturnSelf();
			$flow->expects($this->once())
				->method('setConditions')
				->with($setConditions)
				->willReturnSelf();
			$flow->expects($this->once())
				->method('setActions')
				->with($actions)
				->willReturnSelf();

			$this->manager->expects($this->once())
				->method('getFlowById')
				->with($ruleId)
				->willReturn($flow);

			if ($validationThrows instanceof \Exception) {
				$controller->expects($this->once())
					->method('validateWorkFlow')
					->with($flow)
					->willThrowException($validationThrows);
				$this->manager->expects($this->never())
					->method('updateFlow');
			} else {
				$controller->expects($this->once())
					->method('validateWorkFlow')
					->with($flow);

				$this->manager->expects($this->once())
					->method('updateFlow')
					->with($flow);
				$flow->expects($this->once())
					->method('getId')
					->willReturn($ruleId);
				$this->manager->expects($this->once())
					->method('flowToArray')
					->with($flow)
					->willReturnCallback(function (Flow $flow) {
						return ['flowData' => 'Flow#' . $flow->getId()];
					});
			}
		}

		$response = $controller->updateWorkFlow($ruleId, $name, $conditions, $actions);

		$this->assertInstanceOf('OCP\AppFramework\Http\JSONResponse', $response);
		$this->assertEquals($expectedStatus, $response->getStatus());
		$this->assertEquals($expectedData, $response->getData());
	}

	public function dataDeleteWorkFlow() {
		return [
			[21, false],
			[23, true],
		];
	}

	/**
	 * @dataProvider dataDeleteWorkFlow
	 * @param int $ruleId
	 * @param bool $managerThrows
	 */
	public function testDeleteWorkFlow($ruleId, $managerThrows) {
		$controller = $this->getController();

		if ($managerThrows) {
			$this->manager->expects($this->once())
				->method('getFlowById')
				->with($ruleId)
				->willThrowException(new FlowDoesNotExist());
		} else {
			$flow = $this->getMockBuilder('OCA\Workflow\Engine\Flow')
				->disableOriginalConstructor()
				->getMock();

			$this->manager->expects($this->once())
				->method('getFlowById')
				->with($ruleId)
				->willReturn($flow);
			$this->manager->expects($this->once())
				->method('deleteFlow')
				->with($flow);
		}

		$response = $controller->deleteWorkFlow($ruleId);

		$this->assertInstanceOf('OCP\AppFramework\Http\Response', $response);
		$this->assertEquals(Http::STATUS_OK, $response->getStatus());
	}

	public function dataValidate() {
		return [
			['type', 'flow1', [], ['a']],
			['type', 'flow2', [], ['a']],
			[\str_repeat('b', 64), \str_repeat('a', 200), [], ['a']],
		];
	}

	/**
	 * @dataProvider dataValidate
	 * @param string $type
	 * @param string $name
	 * @param array $conditions
	 * @param array $actions
	 */
	public function testValidate($type, $name, array $conditions, array $actions) {
		$flow = new Flow();
		$flow->setType($type)
			->setName($name)
			->setConditions($conditions)
			->setActions($actions);

		$this->dispatcher->expects($this->once())
			->method('dispatch')
			->with('OCA\Workflow\Engine::validateFlow', $this->anything());

		$controller = $this->getController();
		$this->invokePrivate($controller, 'validateWorkFlow', [$flow]);
	}

	public function dataValidateInvalidCondition() {
		return [
			[[[]], null, null, null],
			[[['rule' => 'rule', 'operator' => 'equals', 'value' => 'true']], new \InvalidArgumentException(), null, null],
			[[['rule' => 'rule', 'operator' => 'equals', 'value' => 'true']], null, new \OutOfBoundsException(), null],
			[[['rule' => 'rule', 'operator' => 'equals', 'value' => 'true']], null, null, new \OutOfBoundsException()],
		];
	}

	/**
	 * @dataProvider dataValidateInvalidCondition
	 * @param array $conditions
	 * @param \Exception|null $getThrows
	 * @param \Exception|null $operatorThrows
	 * @param \Exception|null $valueThrows
	 */
	public function testValidateInvalidCondition($conditions, $getThrows, $operatorThrows, $valueThrows) {
		$rule = $this->getMockBuilder('OCA\Workflow\Tests\Unit\Mock\NoAbstractRule')
			->disableOriginalConstructor()
			->getMock();

		if ($getThrows instanceof \Exception) {
			$this->ruleFactory->expects($this->once())
				->method('getRuleInstance')
				->willThrowException($getThrows);
		} elseif (!empty($conditions[0])) {
			$this->ruleFactory->expects($this->once())
				->method('getRuleInstance')
				->willReturn($rule);
		}

		if ($operatorThrows instanceof \Exception) {
			$rule->expects($this->once())
				->method('validateOperator')
				->willThrowException($operatorThrows);
		}
		if ($valueThrows instanceof \Exception) {
			$rule->expects($this->once())
				->method('validateRuleValue')
				->willThrowException($valueThrows);
		}

		$this->expectException('OutOfBoundsException');

		$flow = new Flow();
		$flow->setType('type')
			->setName('name')
			->setConditions($conditions)
			->setActions([]);
		$this->dispatcher->expects($this->never())
			->method('dispatch');

		$controller = $this->getController();
		$this->invokePrivate($controller, 'validateWorkFlow', [$flow]);
	}

	public function dataValidateInvalid() {
		return [
			[\str_repeat('a', 65), 'name', [], null, 'Workflow type too long or invalid', 1],
			['type', \str_repeat('a', 201), [], null, 'Workflow name too long or invalid', 2],
			['type', 'name', [], true, 'No Autotagging tag given', 3],
		];
	}

	/**
	 * @dataProvider dataValidateInvalid
	 *
	 * @param string $type
	 * @param string $name
	 * @param array $conditions
	 * @param bool $dispatchThrows
	 * @param string $exception
	 * @param int $code
	 */
	public function testValidateInvalid($type, $name, $conditions, $dispatchThrows, $exception, $code) {
		$this->expectException('OutOfBoundsException', $exception, $code);

		$flow = new Flow();
		$flow->setType($type)
			->setName($name)
			->setConditions($conditions)
			->setActions([]);

		if (!$dispatchThrows) {
			$this->dispatcher->expects($this->never())
				->method('dispatch');
		} else {
			$this->dispatcher->expects($this->once())
				->method('dispatch')
				->with('OCA\Workflow\Engine::validateFlow', $this->anything())
				->willThrowException(new \OutOfBoundsException($exception, $code));
		}

		$controller = $this->getController();
		$this->invokePrivate($controller, 'validateWorkFlow', [$flow]);
		$this->assertTrue(true);
	}
}
