<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Retention;

use OCA\Workflow\Engine\Flow;
use OCA\Workflow\Retention\Exception\TagHasNoRetention;
use OCA\Workflow\Retention\Manager;
use OCA\Workflow\Tests\Unit\TestCase;

class ManagerTest extends TestCase {

	/** @var \OCA\Workflow\Engine\FlowManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $flowManager;

	public function setUp(): void {
		parent::setUp();

		$this->flowManager = $this->getMockBuilder('OCA\Workflow\Engine\FlowManager')
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return Manager|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getManager(array $methods = []) {
		if (empty($methods)) {
			return new Manager(
				$this->flowManager
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\Retention\Manager')
				->setConstructorArgs([
					$this->flowManager
				])
				->setMethods($methods)
				->getMock();
		}
	}

	/**
	 * @param array $conditions
	 * @return \OCA\Workflow\Engine\Flow|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getFlowMock(array $conditions = []) {
		$flow = $this->getMockBuilder('OCA\Workflow\Engine\Flow')
			->disableOriginalConstructor()
			->getMock();

		if (!empty($conditions)) {
			$flow->expects($this->any())
				->method('getConditions')
				->willReturn($conditions);
		}

		return $flow;
	}

	public function dataAdd() {
		return [
			['tag1', 10, 'days', null],
			['tag2', 21, 'days', new TagHasNoRetention()],
			['tag3', 23, 'months', new TagHasNoRetention()],
		];
	}

	/**
	 * @dataProvider dataAdd
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 * @param mixed $getThrows
	 */
	public function testAdd($tagId, $numUnits, $unit, $getThrows) {
		$manager = $this->getManager(['get']);

		if ($getThrows instanceof \Exception) {
			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willThrowException($getThrows);

			$this->flowManager->expects($this->once())
				->method('createFlow')
				->willReturn(new Flow());

			$this->flowManager->expects($this->once())
				->method('addFlow')
				->willReturnCallback(function (Flow $flow) use ($tagId, $numUnits, $unit) {
					$this->assertSame('Retention for tag: ' . $tagId, $flow->getName());
					$this->assertSame('retention', $flow->getType());
					$this->assertSame([
						'tagId' => $tagId,
						'numUnits' => $numUnits,
						'unit' => $unit,
					], $flow->getConditions());
					$this->assertSame(['retention' => true], $flow->getActions());
				});
		} else {
			$flow = $this->getFlowMock();

			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willReturn($flow);

			$this->flowManager->expects($this->never())
				->method('addFlow');

			$this->flowManager->expects($this->never())
				->method('createFlow');

			$this->expectException('OCA\Workflow\Retention\Exception\TagAlreadyHasRetention');
		}

		$this->assertEquals([
			'tagId' => $tagId,
			'numUnits' => $numUnits,
			'unit' => $unit,
		], $manager->add($tagId, $numUnits, $unit));
	}

	public function dataUpdate() {
		return [
			['tag1', 10, 'days', null],
			['tag2', 23, 'months', null],
			['tag3', 23, 'months', new TagHasNoRetention()],
		];
	}

	/**
	 * @dataProvider dataUpdate
	 * @param string $tagId
	 * @param int $numUnits
	 * @param string $unit
	 * @param mixed $getThrows
	 */
	public function testUpdate($tagId, $numUnits, $unit, $getThrows) {
		$manager = $this->getManager(['get']);

		if ($getThrows instanceof \Exception) {
			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willThrowException($getThrows);

			$this->flowManager->expects($this->never())
				->method('updateFlow');

			$this->expectException('OCA\Workflow\Retention\Exception\TagHasNoRetention');
		} else {
			$flow = $this->getFlowMock();

			$flow->expects($this->once())
				->method('setConditions')
				->with([
					'tagId' => $tagId,
					'numUnits' => $numUnits,
					'unit' => $unit,
				]);

			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willReturn($flow);

			$this->flowManager->expects($this->once())
				->method('updateFlow')
				->with($flow);
		}

		$this->assertEquals([
			'tagId' => $tagId,
			'numUnits' => $numUnits,
			'unit' => $unit,
		], $manager->update($tagId, $numUnits, $unit));
	}

	public function dataDelete() {
		return [
			['tag1', null],
			['tag2', new TagHasNoRetention()]
		];
	}

	/**
	 * @dataProvider dataDelete
	 * @param string $tagId
	 * @param mixed $getThrows
	 */
	public function testDelete($tagId, $getThrows) {
		$manager = $this->getManager(['get']);

		if ($getThrows instanceof \Exception) {
			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willThrowException($getThrows);

			$this->flowManager->expects($this->never())
				->method('deleteFlow');

			$this->expectException('OCA\Workflow\Retention\Exception\TagHasNoRetention');
		} else {
			$flow = $this->getFlowMock();
			$manager->expects($this->once())
				->method('get')
				->with($tagId)
				->willReturn($flow);

			$this->flowManager->expects($this->once())
				->method('deleteFlow')
				->with($flow);
		}

		$manager->delete($tagId);
	}

	public function dataGet() {
		$flow1 = $this->getFlowMock(['tagId' => 'tag1']);
		$flow2 = $this->getFlowMock(['tagId' => 'tag2']);
		$flow3 = $this->getFlowMock(['tagId' => 'tag3']);
		$flow4 = $this->getFlowMock(['tagId' => 'tag4']);

		return [
			['tag1', [$flow1, $flow2], $flow1],
			['tag4', [$flow3, $flow4], $flow4],
		];
	}

	/**
	 * @dataProvider dataGet
	 * @param string $tagId
	 * @param Flow[] $return
	 * @param Flow $expected
	 */
	public function testGet($tagId, array $return, $expected) {
		$this->flowManager->expects($this->once())
			->method('getRetentionFlows')
			->willReturn($return);

		$manager = $this->getManager();
		$this->assertSame($expected, $this->invokePrivate($manager, 'get', [$tagId]));
	}

	public function dataGetInvalid() {
		$flow1 = $this->getFlowMock(['tagId' => 'tag1']);
		$flow2 = $this->getFlowMock(['tagId' => 'tag2']);

		return [
			['tag3', [$flow1, $flow2]],
			['tag2', []],
		];
	}

	/**
	 * @dataProvider dataGetInvalid
	 * @param string $tagId
	 * @param Flow[] $return
	 */
	public function testGetInvalid($tagId, array $return) {
		$this->expectException(\OCA\Workflow\Retention\Exception\TagHasNoRetention::class);

		$this->flowManager->expects($this->once())
			->method('getRetentionFlows')
			->willReturn($return);

		$manager = $this->getManager();
		$this->invokePrivate($manager, 'get', [$tagId]);
	}

	public function dataGetRetention() {
		$flow1 = $this->getFlowMock(['tagId' => 'tag1']);
		$flow2 = $this->getFlowMock(['tagId' => 'tag2']);

		return [
			['tag1', $flow1, ['tagId' => 'tag1']],
			['tag2', $flow2, ['tagId' => 'tag2']],
		];
	}

	/**
	 * @dataProvider dataGetRetention
	 * @param string $tagId
	 * @param Flow $return
	 * @param array $expected
	 */
	public function testGetRetention($tagId, $return, array $expected) {
		$manager = $this->getManager(['get']);
		$manager->expects($this->once())
			->method('get')
			->willReturn($return);

		$this->assertSame($expected, $manager->getRetention($tagId));
	}

	/**
	 */
	public function testGetRetentionInvalid() {
		$this->expectException(\OCA\Workflow\Retention\Exception\TagHasNoRetention::class);

		$manager = $this->getManager(['get']);
		$manager->expects($this->once())
			->method('get')
			->willThrowException(new TagHasNoRetention());
		$manager->getRetention('tag');
	}

	public function dataGetAll() {
		$flow1 = $this->getFlowMock(['tagId' => 'tag1']);
		$flow2 = $this->getFlowMock(['tagId' => 'tag2']);
		$flow3 = $this->getFlowMock(['tagId' => 'tag3']);

		return [
			[[$flow1, $flow2], [['tagId' => 'tag1'], ['tagId' => 'tag2']]],
			[[$flow3], [['tagId' => 'tag3']]],
			[[], []],
		];
	}

	/**
	 * @dataProvider dataGetAll
	 * @param Flow[] $return
	 * @param array $expected
	 */
	public function testGetAll(array $return, array $expected) {
		$this->flowManager->expects($this->once())
			->method('getRetentionFlows')
			->willReturn($return);

		$manager = $this->getManager();
		$this->assertEquals($expected, $manager->getAll());
	}
}
