<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Retention;

use OCA\Workflow\Retention\MetaBackgroundJob;
use OCA\Workflow\Tests\Unit\TestCase;

/**
 * Class MetaBackgroundJobTest
 *
 * @group DB
 * @package OCA\Workflow\Tests\Retention
 */
class MetaBackgroundJobTest extends TestCase {

	/** @var \OCP\Files\IRootFolder|\PHPUnit\Framework\MockObject\MockObject */
	protected $rootFolder;

	/** @var \OCP\SystemTag\ISystemTagObjectMapper|\PHPUnit\Framework\MockObject\MockObject */
	protected $tagObjectMapper;

	/** @var \OCP\Files\Config\IMountProviderCollection|\PHPUnit\Framework\MockObject\MockObject */
	protected $mountProviderCollection;

	/** @var \OCP\AppFramework\Utility\ITimeFactory|\PHPUnit\Framework\MockObject\MockObject */
	protected $timeFactory;

	/** @var \OCP\ILogger|\PHPUnit\Framework\MockObject\MockObject */
	protected $logger;

	/** @var \OCP\BackgroundJob\IJobList|\PHPUnit\Framework\MockObject\MockObject */
	protected $jobList;

	/** @var \OCP\IConfig|\PHPUnit\Framework\MockObject\MockObject */
	protected $config;

	/** @var \OCP\IUserManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $userManager;

	/** @var \OCA\Workflow\Retention\Manager|\PHPUnit\Framework\MockObject\MockObject */
	protected $retentionManager;

	public function setUp(): void {
		parent::setUp();

		$this->rootFolder = $this->getMockBuilder('OCP\Files\IRootFolder')
			->disableOriginalConstructor()
			->getMock();

		$this->tagObjectMapper = $this->getMockBuilder('OCP\SystemTag\ISystemTagObjectMapper')
			->disableOriginalConstructor()
			->getMock();

		$this->mountProviderCollection = $this->getMockBuilder('OCP\Files\Config\IMountProviderCollection')
			->disableOriginalConstructor()
			->getMock();

		$this->timeFactory = $this->getMockBuilder('OCP\AppFramework\Utility\ITimeFactory')
			->disableOriginalConstructor()
			->getMock();

		$this->config = $this->getMockBuilder('OCP\IConfig')
			->disableOriginalConstructor()
			->getMock();

		$this->userManager = $this->getMockBuilder('OCP\IUserManager')
			->disableOriginalConstructor()
			->getMock();

		$this->logger = $this->getMockBuilder('OCP\ILogger')
			->disableOriginalConstructor()
			->getMock();

		$this->jobList = $this->getMockBuilder('OCP\BackgroundJob\IJobList')
			->disableOriginalConstructor()
			->getMock();

		$this->retentionManager = $this->getMockBuilder('OCA\Workflow\Retention\Manager')
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return MetaBackgroundJob|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getJob(array $methods = []) {
		if (empty($methods)) {
			return new MetaBackgroundJob(
				$this->jobList,
				$this->logger,
				$this->config,
				$this->userManager,
				$this->retentionManager
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\Retention\MetaBackgroundJob')
				->setConstructorArgs([
					$this->jobList,
					$this->logger,
					$this->config,
					$this->userManager,
					$this->retentionManager
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function testConstruct() {
		$job = $this->getJob();
		$this->assertInstanceOf('OC\BackgroundJob\TimedJob', $job);
		$this->assertInstanceOf('OCP\BackgroundJob\IJob', $job);

		$this->assertInstanceOf('OCP\ILogger', $this->invokePrivate($job, 'logger'));
		$this->assertInstanceOf('OCP\BackgroundJob\IJobList', $this->invokePrivate($job, 'jobList'));
		$this->assertInstanceOf('OCA\Workflow\Retention\Manager', $this->invokePrivate($job, 'retentionManager'));
	}

	public function testConstructNoDi() {
		$job = new MetaBackgroundJob();
		$this->assertInstanceOf('OC\BackgroundJob\TimedJob', $job);
		$this->assertInstanceOf('OCP\BackgroundJob\IJob', $job);

		$this->assertInstanceOf('OCP\ILogger', $this->invokePrivate($job, 'logger'));
		$this->assertInstanceOf('OCP\BackgroundJob\IJobList', $this->invokePrivate($job, 'jobList'));
		$this->assertInstanceOf('OCA\Workflow\Retention\Manager', $this->invokePrivate($job, 'retentionManager'));
	}

	public function dataRun() {
		return [
			['tagbased', true, false, false],
			['userbased', false, true, false],
			['invalid', false, false, true],
		];
	}

	/**
	 * @dataProvider dataRun
	 * @param string $engine
	 * @param bool $executeTags
	 * @param bool $executeUsers
	 * @param bool $logCritical
	 */
	public function testRun($engine, $executeTags, $executeUsers, $logCritical) {
		$job = $this->getJob([
			'createTagJobs',
			'createUserJobs',
		]);

		$this->config->expects($this->once())
			->method('getSystemValue')
			->with('workflow.retention_engine', 'tagbased')
			->willReturn($engine);

		if ($executeTags) {
			$job->expects($this->once())
				->method('createTagJobs');
			$job->expects($this->never())
				->method('createUserJobs');

			$this->jobList->expects($this->once())
				->method('remove')
				->with('\OCA\Workflow\Retention\UserBasedRetention');
		} elseif ($executeUsers) {
			$job->expects($this->once())
				->method('createUserJobs');
			$job->expects($this->never())
				->method('createTagJobs');

			$this->jobList->expects($this->once())
				->method('remove')
				->with('\OCA\Workflow\Retention\TagBasedRetention');
		} elseif ($executeTags) {
			$job->expects($this->never())
				->method('createTagJobs');
			$job->expects($this->never())
				->method('createUserJobs');
			$this->jobList->expects($this->never())
				->method('remove');
		}

		$this->logger->expects($logCritical ? $this->once() : $this->never())
			->method('critical');

		$this->invokePrivate($job, 'run', [null]);
	}

	public function testCreateTagJobs() {
		$job = $this->getJob();

		$this->retentionManager->expects($this->once())
			->method('getAll')
			->willReturn([
				['tagId' => '23'],
				['tagId' => '42'],
			]);

		$this->jobList->expects($this->exactly(2))
			->method('has')
			->willReturnMap([
				['\OCA\Workflow\Retention\TagBasedRetention', ['tagId' => '23'], false],
				['\OCA\Workflow\Retention\TagBasedRetention', ['tagId' => '42'], true],
			]);

		$this->jobList->expects($this->once())
			->method('add')
			->with('\OCA\Workflow\Retention\TagBasedRetention', ['tagId' => '23']);

		$this->logger->expects($this->once())
			->method('warning')
			->with('Retention job for tag "{tag}" was missing.', [
				'tag' => '23',
				'app' => 'workflow',
			]);

		$this->invokePrivate($job, 'createTagJobs');
	}

	public function testCreateUserJobs() {
		$job = $this->getJob(['createClosureForUserIteration']);

		$closure = function () {
		};
		$job->expects($this->once())
			->method('createClosureForUserIteration')
			->willReturn($closure);

		$this->userManager->expects($this->once())
			->method('callForAllUsers')
			->with($closure);

		$this->invokePrivate($job, 'createUserJobs');
	}

	public function testCreateClosureForUserIteration() {
		$job = $this->getJob();
		/** @var \Closure $closure */
		$closure = $this->invokePrivate($job, 'createClosureForUserIteration', [$this->jobList]);

		$user = $this->createMock(\OCP\IUser::class);
		$user->expects($this->once())
			->method('getUID')
			->willReturn('uid');

		$this->jobList->expects($this->once())
			->method('add')
			->with('\OCA\Workflow\Retention\UserBasedRetention', ['user' => 'uid']);

		$closure(null);
		$closure($user);
	}
}
