<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Retention;

use Exception;
use OCA\Workflow\Retention\Exception\TagHasNoRetention;
use OCA\Workflow\Retention\Manager;
use OCA\Workflow\Retention\RetentionChecker;
use OCA\Workflow\Retention\TagBasedRetention;
use OCA\Workflow\Tests\Unit\TestCase;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\BackgroundJob\IJobList;
use OCP\Files\Config\IMountProviderCollection;
use OCP\Files\IRootFolder;
use OCP\IConfig;
use OCP\ILogger;
use OCP\IUserSession;
use OCP\SystemTag\ISystemTagObjectMapper;
use OCP\SystemTag\TagNotFoundException;
use OCP\IUser;
use OCP\Files\Config\IUserMountCache;
use OCP\Files\Config\ICachedMountInfo;
use OCP\BackgroundJob\IJob;
use OC\BackgroundJob\TimedJob;
use PHPUnit\Framework\MockObject\MockObject;

/**
 * Class TagBasedRetentionTest
 *
 * @group DB
 * @package OCA\Workflow\Tests\Retention
 */
class TagBasedRetentionTest extends TestCase {

	/** @var IRootFolder|MockObject */
	protected $rootFolder;

	/** @var ISystemTagObjectMapper|MockObject */
	protected $tagObjectMapper;

	/** @var IMountProviderCollection|MockObject */
	protected $mountProviderCollection;

	/** @var ITimeFactory|MockObject */
	protected $timeFactory;

	/** @var IUserSession|MockObject */
	protected $userSession;

	/** @var ILogger|MockObject */
	protected $logger;

	/** @var IJobList|MockObject */
	protected $jobList;

	/** @var Manager|MockObject */
	protected $retentionManager;
	/**
	 * @var IConfig|MockObject
	 */
	private $config;
	/**
	 * @var RetentionChecker|MockObject
	 */
	private $retentionChecker;

	public function setUp(): void {
		parent::setUp();

		$this->rootFolder = $this->getMockBuilder(IRootFolder::class)
			->disableOriginalConstructor()
			->getMock();

		$this->tagObjectMapper = $this->getMockBuilder(ISystemTagObjectMapper::class)
			->disableOriginalConstructor()
			->getMock();

		$this->mountProviderCollection = $this->getMockBuilder(IMountProviderCollection::class)
			->disableOriginalConstructor()
			->getMock();

		$this->timeFactory = $this->getMockBuilder(ITimeFactory::class)
			->disableOriginalConstructor()
			->getMock();

		$this->userSession = $this->getMockBuilder(IUserSession::class)
			->disableOriginalConstructor()
			->getMock();

		$this->logger = $this->getMockBuilder(ILogger::class)
			->disableOriginalConstructor()
			->getMock();

		$this->jobList = $this->getMockBuilder(IJobList::class)
			->disableOriginalConstructor()
			->getMock();

		$this->retentionManager = $this->getMockBuilder(Manager::class)
			->disableOriginalConstructor()
			->getMock();

		$this->config = $this->createMock(IConfig::class);
		$this->retentionChecker = $this->createMock(RetentionChecker::class);
	}

	/**
	 * @param array $methods
	 * @return TagBasedRetention|MockObject
	 */
	protected function getJob(array $methods = []) {
		if (empty($methods)) {
			return new TagBasedRetention(
				$this->rootFolder,
				$this->tagObjectMapper,
				$this->mountProviderCollection,
				$this->timeFactory,
				$this->userSession,
				$this->logger,
				$this->jobList,
				$this->retentionManager,
				$this->config,
				$this->retentionChecker
			);
		}

		return $this->getMockBuilder(TagBasedRetention::class)
			->setConstructorArgs([
				$this->rootFolder,
				$this->tagObjectMapper,
				$this->mountProviderCollection,
				$this->timeFactory,
				$this->userSession,
				$this->logger,
				$this->jobList,
				$this->retentionManager,
				$this->config,
				$this->retentionChecker
			])
			->setMethods($methods)
			->getMock();
	}

	public function testConstruct() {
		$job = $this->getJob();
		$this->assertInstanceOf(TimedJob::class, $job);
		$this->assertInstanceOf(IJob::class, $job);

		$this->assertInstanceOf(IRootFolder::class, self::invokePrivate($job, 'rootFolder'));
		$this->assertInstanceOf(ISystemTagObjectMapper::class, self::invokePrivate($job, 'tagObjectMapper'));
		$this->assertInstanceOf(IMountProviderCollection::class, self::invokePrivate($job, 'mountProviderCollection'));
		$this->assertInstanceOf(Manager::class, self::invokePrivate($job, 'retentionManager'));
		$this->assertInstanceOf(ITimeFactory::class, self::invokePrivate($job, 'timeFactory'));
		$this->assertInstanceOf(IUserSession::class, self::invokePrivate($job, 'userSession'));
		$this->assertInstanceOf(ILogger::class, self::invokePrivate($job, 'logger'));
		$this->assertInstanceOf(IJobList::class, self::invokePrivate($job, 'jobList'));
	}

	public function testConstructNoDi() {
		$job = new TagBasedRetention();
		$this->assertInstanceOf(TimedJob::class, $job);
		$this->assertInstanceOf(IJob::class, $job);

		$this->assertInstanceOf(IRootFolder::class, self::invokePrivate($job, 'rootFolder'));
		$this->assertInstanceOf(ISystemTagObjectMapper::class, self::invokePrivate($job, 'tagObjectMapper'));
		$this->assertInstanceOf(IMountProviderCollection::class, self::invokePrivate($job, 'mountProviderCollection'));
		$this->assertInstanceOf(Manager::class, self::invokePrivate($job, 'retentionManager'));
		$this->assertInstanceOf(ITimeFactory::class, self::invokePrivate($job, 'timeFactory'));
		$this->assertInstanceOf(IUserSession::class, self::invokePrivate($job, 'userSession'));
		$this->assertInstanceOf(ILogger::class, self::invokePrivate($job, 'logger'));
		$this->assertInstanceOf(IJobList::class, self::invokePrivate($job, 'jobList'));
	}

	public function testGetUserFolder() {
		$job = $this->getJob();

		$user1 = $this->createMock(IUser::class);
		$user1->expects($this->atLeastOnce())
			->method('getUID')
			->willReturn('user1');

		$admin = $this->createMock(IUser::class);
		$admin->expects($this->atLeastOnce())
			->method('getUID')
			->willReturn('admin');

		$this->rootFolder->expects($this->exactly(2))
			->method('getUserFolder')
			->withConsecutive(
				['admin'],
				['user1']
			)
			->willReturnMap([
				['admin', 'admin-folder'],
				['user1', 'user1-folder'],
			]);

		$this->assertSame('admin-folder', self::invokePrivate($job, 'getUserFolder', [$admin]));
		$this->assertSame('admin-folder', self::invokePrivate($job, 'getUserFolder', [$admin]));

		$this->assertSame('user1-folder', self::invokePrivate($job, 'getUserFolder', [$user1]));
	}

	public function testRun() {
		$job = $this->getJob([
			'executeRetention',
		]);

		$job->expects($this->once())
			->method('executeRetention')
			->with(23);

		self::invokePrivate($job, 'run', [['tagId' => 23]]);
	}

	public function dataExecuteRetention(): array {
		return [
			['tag', new TagHasNoRetention(), 0, 0, 0, 1, [], false],
			['tag', ['tagId' => 'tag', 'numUnits' => -1, 'unit' => 'dayz'], 0, 0, 0, 1, [], false],
			['tag2', ['tagId' => 'tag2', 'numUnits' => 1, 'unit' => 'days'], 1, 0, 0, 0, [
				0 => [],
			], true],
			['tag2', ['tagId' => 'tag2', 'numUnits' => 1, 'unit' => 'days'], 2, 1, 1, 0, [
				0 => [23],
				1 => [],
			], true],
			['tag2', ['tagId' => 'tag2', 'numUnits' => 1, 'unit' => 'days'], 1, 0, 0, 1, [
				0 => new TagNotFoundException(),
			], false],
			['tag2', ['tagId' => 'tag2', 'numUnits' => 1, 'unit' => 'days'], 2, 1, 1, 1, [
				0 => [23],
				1 => new TagNotFoundException(),
			], false],
		];
	}

	protected function getNodeMock($class = 'Node'): MockObject {
		return $this->getMockBuilder('OCP\\Files\\' . $class)
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @dataProvider dataExecuteRetention
	 * @param string $tagId
	 * @param array $retention
	 * @param int $numTagCalls
	 * @param int $numUserFolderCalls
	 * @param int $numRetentionCalls
	 * @param int $numLoggerCalls
	 * @param $objectsReturned
	 * @param $expected
	 */
	public function testExecuteRetention($tagId, $retention, $numTagCalls, $numUserFolderCalls, $numRetentionCalls, $numLoggerCalls, $objectsReturned, $expected) {
		$job = $this->getJob([
			'getUserFolder',
			'checkItemsForRetention',
			'initFilesystemForUser',
		]);

		$idItems = [23 => [
			$this->getNodeMock('Folder'),
			$this->getNodeMock('File'),
		]];

		$folder = $this->getNodeMock('Folder');
		$folder->expects($this->exactly($numUserFolderCalls))
			->method('getById')
			->willReturnMap([
				[23, false, $idItems[23]]
			]);

		if ($retention instanceof Exception) {
			$this->retentionManager->expects($this->once())
				->method('getRetention')
				->with($tagId)
				->willThrowException($retention);
		} else {
			$this->retentionManager->expects($this->once())
				->method('getRetention')
				->with($tagId)
				->willReturn($retention);
		}

		$user = $this->createMock(IUser::class);

		$job->expects($this->exactly($numUserFolderCalls))
			->method('initFilesystemForUser')
			->with($user);
		$job->expects($this->exactly($numUserFolderCalls))
			->method('getUserFolder')
			->willReturnMap([
				[$user, $folder]
			]);

		$cacheInfo = $this->getMockBuilder(ICachedMountInfo::class)
			->disableOriginalConstructor()
			->getMock();
		$cacheInfo->expects($this->exactly($numUserFolderCalls))
			->method('getUser')
			->willReturn($user);

		$mountCache = $this->getMockBuilder(IUserMountCache::class)
			->disableOriginalConstructor()
			->getMock();
		$mountCache->expects($this->exactly($numUserFolderCalls))
			->method('getMountsForFileId')
			->willReturn([1 => $cacheInfo]);
		if (!($retention instanceof Exception)) {
			$this->mountProviderCollection->expects($this->once())
				->method('getMountCache')
				->willReturn($mountCache);
		}

		$this->logger->expects($this->exactly($numLoggerCalls))
			->method('error');

		global $expectedNodes, $callCount;
		$callCount = -1;
		$expectedNodes = [];
		$this->tagObjectMapper->expects($this->exactly($numTagCalls))
			->method('getObjectIdsForTags')
			->willReturnCallback(static function () use ($objectsReturned, $idItems, $user, &$expectedNodes, &$callCount) {
				$callCount++;
				if ($objectsReturned[$callCount] instanceof Exception) {
					throw $objectsReturned[$callCount];
				}

				if (empty($objectsReturned[$callCount])) {
					return [];
				}

				$expectedNodes = [['node' => $idItems[$objectsReturned[$callCount][0]][0], 'owner' => $user]];

				return $objectsReturned[$callCount];
			});

		$job->expects($this->exactly($numRetentionCalls))
			->method('checkItemsForRetention')
			->willReturnCallback(function ($time, $nodes) {
				global $expectedNodes;
				$this->assertIsInt($time);
				$this->assertSame($expectedNodes, $nodes);
			});

		$this->assertSame($expected, self::invokePrivate($job, 'executeRetention', [$tagId]));
	}

	public function testCheckItemsForRetention() {
		$job = $this->getJob();

		// File, mtime < so deleted
		$file = $this->getNodeMock('File');
		$file
			->method('getId')
			->willReturn(1);
		$file
			->method('getMTime')
			->willReturn(122);
		$file->expects($this->once())
			->method('getOwner')
			->willReturn($this->createMock(IUser::class));
		$file->expects($this->once())
			->method('delete');

		// File that throws an exception on deletion
		$fileThrows = $this->getNodeMock('File');
		$fileThrows
			->method('getId')
			->willReturn(2);
		$fileThrows
			->method('getMTime')
			->willReturn(122);
		$fileThrows->expects($this->once())
			->method('getOwner')
			->willReturn($this->createMock(IUser::class));
		$fileThrows->expects($this->once())
			->method('delete')
			->willThrowException(new Exception());

		// File that throws an exception on getId
		$fileIdThrows = $this->getNodeMock('File');
		$fileIdThrows
			->method('getId')
			->willThrowException(new Exception());

		// Root folder, mtime > so not deleted
		$folder = $this->getNodeMock('Folder');
		$folder
			->method('getId')
			->willReturn(4);
		$folder
			->method('getMTime')
			->willReturn(124);
		$folder->expects($this->once())
			->method('getDirectoryListing')
			->willReturn([$fileThrows, $fileIdThrows, $file]);
		$folder->expects($this->never())
			->method('delete');

		// Folder that throws an exception on getDirectoryListing
		$folderThrows = $this->getNodeMock('Folder');
		$folderThrows
			->method('getId')
			->willReturn(5);
		$folderThrows
			->method('getMTime')
			->willReturn(124);
		$folderThrows->expects($this->once())
			->method('getDirectoryListing')
			->willThrowException(new Exception());
		$folderThrows->expects($this->never())
			->method('delete');

		// File that throws an exception on getId
		$folderIdThrows = $this->getNodeMock('Folder');
		$folderIdThrows
			->method('getId')
			->willThrowException(new Exception());

		// Folder that throws an exception on deletion
		$folderAlreadyChecked = $this->getNodeMock('Folder');
		$folderAlreadyChecked
			->method('getId')
			->willReturn(3);
		$folderAlreadyChecked->expects($this->never())
			->method('getMTime');
		$folderAlreadyChecked->expects($this->never())
			->method('delete');

		$this->retentionChecker->method('isRetentionOver')
			->withConsecutive(
				[$folder, 123], // 4
				[$file, 123], // 1
				[$fileThrows, 123] // 2
			)
			->willReturnOnConsecutiveCalls(false, true, true);

		self::invokePrivate($job, 'checkedNodes', [[3 => true]]);
		self::invokePrivate($job, 'checkItemsForRetention', [
			123,
			[
				['node' => $folderThrows, 'owner' => $this->createMock(IUser::class)],
				['node' => $folderAlreadyChecked, 'owner' => $this->createMock(IUser::class)],
				['node' => $folderIdThrows, 'owner' => $this->createMock(IUser::class)],
				['node' => $folder, 'owner' => $this->createMock(IUser::class)],
			]
		]);
	}

	public function dataGetRetentionTimestamp(): array {
		return [
			[7, 'days', 1414195200], // 2014-10-25T00:00:00+00:00
			[1, 'weeks', 1414195200], // 2014-10-25T00:00:00+00:00
			[3, 'months', 1406851200], // 2014-08-01T00:00:00+00:00
			[20, 'years', 783648000], // 1994-11-01T00:00:00+00:00
		];
	}

	/**
	 * @dataProvider dataGetRetentionTimestamp
	 * @param int $numUnits
	 * @param string $unit
	 * @param int $expected
	 */
	public function testGetRetentionTimestamp($numUnits, $unit, $expected) {
		$job = $this->getJob();

		$this->timeFactory->expects($this->once())
			->method('getTime')
			->willReturn(1414800000);// 2014-11-01T00:00:00+00:00

		$this->assertSame($expected, self::invokePrivate($job, 'getRetentionTimestamp', [[
			'numUnits' => $numUnits,
			'unit' => $unit,
		]]));
	}

	public function dataGetRetentionTimestampInvalid(): array {
		return [
			['7', 'days', 'Invalid retention period for tag "{tag}"', 1],
			['-7', 'days', 'Invalid retention period for tag "{tag}"', 1],
			[-7, 'days', 'Invalid retention period for tag "{tag}"', 1],
			[1, 'hours', 'Invalid retention period unit for tag "{tag}"', 2],
			[3, 'minutes', 'Invalid retention period unit for tag "{tag}"', 2],
			[20, 'seconds', 'Invalid retention period unit for tag "{tag}"', 2],
		];
	}

	/**
	 * @dataProvider dataGetRetentionTimestampInvalid
	 * @param int $numUnits
	 * @param string $unit
	 * @param string $message
	 * @param int $code
	 */
	public function testGetRetentionTimestampInvalid($numUnits, $unit, $message, $code) {
		$job = $this->getJob();

		$this->timeFactory->expects($this->never())
			->method('getTime');

		$this->expectException('OutOfBoundsException');
		$this->expectExceptionMessage($message);
		$this->expectExceptionCode($code);

		self::invokePrivate($job, 'getRetentionTimestamp', [[
			'numUnits' => $numUnits,
			'unit' => $unit,
		]]);
	}
}
